 /* @(#) $Id: saprfc.h,v 1.1.1.1 2006/03/07 08:56:42 rosenauer Exp $ SAP*/


/*---------------------------------------------------------------------------------
 * R F C  S D K
 *
 *
 *  (c) Copyright SAP AG, Walldorf  1993 - 1999
 *
 *	Portion of this software is based on the work from
 *	Visual Edge Software
 *---------------------------------------------------------------------------------
 */

/*---------------------------------------------------------------------------------
 * internal hint: this file is shipped to customers and must not include or use
 * any SAP internal data types or include files.
 *---------------------------------------------------------------------------------
 */

/*
 * @doc
 *
 * @module Remote Function Call API  |
 *
 * The Remote Function Call API (RFC API) allows -
 * remotely or locally - calling ABAP/4 function
 * modules from C programs as well as receiving call
 * request issued from an ABAP/4 program by the
 * CALL FUNCTION interface.
 *
 * In ABAP/4 function modules are special routines with
 * well defined and documented interfaces, which are
 * developed within a library development workbench.
 *
 * ABAP/4 function modules use named parameter passing.
 * Also there are exceptions which can be raised by a function
 * module to indicate errors. These exceptions can be caught
 * by the caller of a function module.
 *
 * The following functions form the programming interface to
 * call ABAP/4 function modules from a C environment.
 *
 * @group Functions for a RFC client program
 * @flag <f RfcOpenEx>        | opens an RFC connection including SAP logon,
 *                              can also be used for
 *                              non C environments such as Visual Basic.
 *                              This call is recommended instead of RfcOpen,
 *                              RfcOpenExt, RfcOpenExtV3 and RfcConnect.
 * @flag <f RfcOpen>          | opens an RFC connection.
 * @flag <f RfcOpenExt>       | another way to open an RFC connection more
 *                              appropriate for non C environments such as Visual Basic.
 * @flag <f RfcOpenExtV3>     | another way to open an RFC connection more
 *                              appropriate for non C environments such as Visual Basic
 *                              (using RFC Version 3).
 * @flag <f RfcConnect>       | opens an RFC connection via LOAD BALANCING (R/3 only).
 * @flag <f RfcCall>          | Calls an ABAP/4 function module.
 * @flag <f RfcReceive>       | Receives the return values from an
 *                              ABAP/4 function module.
 * @flag <f RfcCallReceive>   | Calling a function module and receiving the
 *                              return values in one step.
 * @flag <f RfcCreateTransID> | Gets a unique Transaction-ID for calling an ABAP/4
 *                              function module using the transactional RFC Interface.
 * @flag <f RfcIndirectCall>  | Calls an ABAP/4 function module using transactional RFC
 *                              Interface in R/3 (implizite confirmation).
 * @flag <f RfcIndirectCallEx> | Calls an ABAP/4 function module using transactional RFC
 *                              Interface in R/3 ( (explizite confirmation).
 * @flag <f RfcConfirmTransID> | Confirms a transaction called by RfcIndirectCallEx.
 * @flag <f RfcCleanupContext> | Cleanup the attached R/3 context
 * @flag <f RfcRegisterPasswordChanger> | Register a function to change expired passwords
 *
 * @flag <f RfcCancel>         | Cancel an rfc call. This function works only with
 *                               RfcCall, RfcListen, RfcReceive construct.
 *
 * @group Functions for a RFC server program
 * @flag <f RfcAccept>                    | accepting a RFC request or registering at a
 *                                          SAP gateway.
 * @flag <f RfcInstallFunction>           | registering functions as callable
 *                                          function modules.
 * @flag <f RfcDispatch>                  | waiting for the next function call.
 * @flag <f RfcGetData>                   | receiving the parameters of a function.
 * @flag <f RfcDynamicGetData>            | receiving the parameters of a function one-by-one.
 * @flag <f RfcSendData>                  | sending back the return values.
 * @flag <f RfcRaise>                     | raising an exception.
 * @flag <f RfcGetName>                   | reading the symbolic function name.
 *                                          (Max. 30 B)
 * @flag <f RfcGetLongName>		  | internal use only. Reading the symbolic function name longer than 30.
 * @flag <f RfcGetNameEx>                 | internal use only. Reading the symbolic function name.
 *                                          (Max. 30 B) and returning an according
 *                                          return code to the RFC program after
 *                                          a system call such as RFC_PING for
 *                                          connection test is automatically executed
 *                                          by the RFC library.
 * @flag <f RfcGetLongNameEx>             | reading the symbolic function name longer than 30.
 *                                          and returning code to the RFC program after
 *                                          a system call such as RFC_PING for
 *                                          connection test is automatically executed
 *                                          by the RFC library.

 * @flag <f RfcCallReceive>               | Calling a function module and receiving
 *                                          the return values in one step. It can be
 *                                          be used for calling back a function module
 *                                          in R/3 using the same RFC connection while
 *                                          executing a RFC function in this server
 *                                          program. Using <f RfcCall> and <f RfcReceive>
 *                                          is also possible.
 * @flag <f RfcInstallTransactionControl> | registering functions as callable
 *                                          function modules for transactional RFC.
 * @flag <f RfcSncMode>                   | getting the <t RFC_SNC_MODE> (ON or OFF)
 * @flag <f RfcSncPartnerName>            | getting the SNC name of the RFC client
 * @flag <f RfcSncPartnerAclKey>          | getting the SNC ACL key of the RFC client
 *
 * @group Special Functions
 * @flag <f RfcListen> 	          | listen for incoming RFC events.
 * @flag <f RfcWaitForRequest>    | wait for incoming RFC requests.
 *                                  Only available after <f RfcAccept> in
 *                                  register mode.
 * @flag <f RfcGetAttributes>     | get some information about this RFC connection
 * @flag <f RfcInstallStructure>  | define a structure for working with inhomogeneous
 *                                  structure or internal table in an RFC client or
 *                                  server program.
 * @flag <f RfcCheckRegisterServer>   | check for registered RFC server at a SAP gateway
 * @flag <f RfcCancelRegisterServer>  | cancel all registered RFC servers at a SAP gateway
 * @flag <f RfcSncNameToAclKey>   | convert an SNC name in an SNC ACL key
 * @flag <f RfcSncAclKeyToName>   | convert an SNC ACL key in an SNC name
 *
 * @flag <f RfcAllowStartProgram> | Allows the RFC library for starting some programs
 *
 * @flag <f RfcSetTrace>          | Activate/Deactivate the RFC trace
 *
 * @flag <f RfcQueueInsert>       | Insert the tRFC-call in the R/3 receive queue
 *                                  instead of immediately executing as RfcIndeirectCall
 *                                  does.
 * @flag <f RfcIsoLangToSap>	  | Converts 2-byte ISO-Language into 1-byte SAP-Language
 * @flag <f RfcSapLangToIso>       | Converts 1-byte SAP-Language into 2-byte ISO-Languge
 *
 * @flag <f RfcCharConversion>    | Turn off/on the conversion of charakters between
 *                                  own codepage and partner's codepage
 *
 * @flag <f RfcIsValidHandle>     | Check handle
 *
 * @flag <f RfcGetTicket>         | Retrieve backend generated cookie version 2 after calling RfcOpenEx
 *                                  with flag GETSSO2
 *
 * @flag <f RfcSetCodePage>       | Set codepage for a rfc connection
 *
 * @flag <f RfcSetSystemCodePage> | Change default codepage for all further connections
 */

#ifndef RFC_H
#define RFC_H

#ifndef SAPRFC_H
#define SAPRFC_H

#ifdef __midl
/*
 * this part is used when compiled via MIDL
 */
#define SAPonNT
#else /* C/C++ */
/* wchar_t */
#include<stdlib.h>
/* integers */
#ifndef INT_MAX
#include<limits.h>
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#define RFC_VERSION 3


/*
 * For Windows 3.x _WINDOWS has to be defined, as it is done
 * by MS Visual C++
 */

#ifndef SAPTYPE_H

#ifdef _WIN32
#  ifndef SAPonNT
#    define SAPonNT
#  endif
#endif  /* _WIN32       */

#ifdef _WINDOWS
#  ifndef _WIN32
#    ifndef SAPonWINDOWS
#      define SAPonWINDOWS
#    endif
#  endif  /* _WIN16     */
#endif  /* _WINDOWS     */

#ifdef __OS400__
#  ifndef SAPonOS400
#    define SAPonOS400
#  endif
#  ifndef SAPonAS400
#    define SAPonAS400
#  endif
#endif  /* __OS400__ */

#endif  /* SAPTYPE_H    */


#endif /* MIDL <-> C/C++ */


#if !defined(SAP_FAR)|| !defined(SAP_PASCAL)|| !defined(SAP_EXPORT)|| !defined(SAP_LOADDS) || !defined(SAP_STDCALL)
#   undef SAP_FAR
#   undef SAP_PASCAL
#   undef SAP_EXPORT
#   undef SAP_LOADDS
#   undef SAP_STDCALL
#   if defined(SAPonWINDOWS)
#       define SAP_FAR __far
#       define SAP_PASCAL __pascal
#       define SAP_EXPORT __export
#       define SAP_LOADDS __loadds
#       define SAP_STDCALL
#   elif defined(SAPonNT)
#       define SAP_FAR
#       define SAP_PASCAL
#       define SAP_EXPORT
#       define SAP_LOADDS
#       define SAP_STDCALL _stdcall
#   else
#       define SAP_FAR
#       define SAP_PASCAL
#       define SAP_EXPORT
#       define SAP_LOADDS
#       define SAP_STDCALL
#   endif
#endif

#if !defined(_SYSTEM)
#   if defined(OS2)
#       define _SYSTEM _System
#   else
#       define _SYSTEM
#       define SAP_API SAP_FAR SAP_LOADDS SAP_PASCAL SAP_STDCALL
#   endif
#endif /* SAP_API */

#if !defined(SAP_API)
#   if defined(OS2)
#       define SAP_API _SYSTEM
#   endif
#endif

#if !defined(WIN_DLL_EXPORT_FLAGS)
#   define WIN_DLL_EXPORT_FLAGS SAP_API
#endif /* WIN_DLL_EXPORT_FLAGS */


#if !defined(DLL_CALL_BACK_FUNCTION)
#   if defined(SAPonWINDOWS)
#       define DLL_CALL_BACK_FUNCTION SAP_FAR SAP_PASCAL
#   elif defined(SAPonNT)
#       define DLL_CALL_BACK_FUNCTION SAP_STDCALL
#   elif defined(OS2)
#       define DLL_CALL_BACK_FUNCTION _SYSTEM
#   else
#       define DLL_CALL_BACK_FUNCTION
#   endif
#   if defined(OS2)
#      define DLL_CALL_BACK_FUNCTION_PTR * _SYSTEM
#   else
#      define DLL_CALL_BACK_FUNCTION_PTR DLL_CALL_BACK_FUNCTION *
#   endif
#endif /* DLL_CALL_BACK_FUNCTION */


#if ! defined (RUNT) && ! defined (ABAP0LIB)

/* ABAP/4 data types */
#define TYPC            0
#define TYPDATE         1
#define TYPP            2
#define TYPTIME         3
#define TYPX            4
#define TYPTABH         5
#define TYPNUM          6
#define TYPFLOAT        7
#define TYPINT          8
#define TYPINT2         9
#define TYPINT1         10
/*  TYPW = 11 */
#define TYP1            12
#define TYP2            13


/* internal hint :
 * ABAP/4 data types must also be defined in abtypes.h
 */

#else
#include "abtypes.h"
#endif /* ! defined (RUNT) && ! defined (ABAP0LIB) */


#ifdef SAPwithUNICODE

#define RfcInit			RfcInitU
#define RfcEnvironment		RfcEnvironmentU
#define RfcLastError		RfcLastErrorU
#define RfcOpenExt		RfcOpenExtU
#define RfcOpen			RfcOpenU
#define RfcAccept		RfcAcceptU
#define RfcAcceptExt		RfcAcceptExtU
#define RfcClose		RfcCloseU
#define RfcAbort 		RfcAbortU
#define RfcCall			RfcCallU
#define RfcReceive		RfcReceiveU
#define RfcCallReceive		RfcCallReceiveU
#define RfcIndirectCall		RfcIndirectCallU
#define RfcGetName		RfcGetNameU
#define RfcGetData		RfcGetDataU
#define RfcDynamicGetData	RfcDynamicGetDataU
#define RfcSendData		RfcSendDataU
#define RfcRaise		RfcRaiseU
#define RfcListen		RfcListenU
#define RfcFileArgv		RfcFileArgvU
#define RfcConnArgv		RfcConnArgvU
#define RfcSendSystemInfo	RfcSendSystemInfoU
#define RfcInstallFunction	RfcInstallFunctionU
#define RfcInstallFunctionExt	RfcInstallFunctionExtU
#define RfcDispatch		RfcDispatchU
#define RfcGetTransID		RfcGetTransIDU
#define RfcCreateTransID                RfcCreateTransIDU
#define RfcCreateLocalTransID           RfcCreateLocalTransIDU
#define RfcGuidToTid                    RfcGuidToTidU
#define RfcTidToGuid                    RfcTidToGuidU
#define RfcCheckTransID		RfcCheckTransIDU
#define RfcConfirmTransID	RfcConfirmTransIDU
#define RfcAllocParamSpace	RfcAllocParamSpaceU
#define RfcFreeParamSpace	RfcFreeParamSpaceU
#define RfcAddExportParam	RfcAddExportParamU
#define RfcAddImportParam	RfcAddImportParamU
#define RfcAddTable		RfcAddTableU
#define RfcGetTableHandle	RfcGetTableHandleU
#define RfcCallExt		RfcCallExtU
#define RfcReceiveExt		RfcReceiveExtU
#define RfcCallReceiveExt	RfcCallReceiveExtU
#define RfcIndirectCallExt	RfcIndirectCallExtU
#define RfcGetDataExt		RfcGetDataExtU
#define RfcSendDataExt		RfcSendDataExtU
#define RfcExecProgram		RfcExecProgramU
#define RfcGetVersion		RfcGetVersionU
#define RfcCheckVersionsOfUsedLibs RfcCheckVersionsOfUsedLibsU
#define RfcGetAllLibVersions	RfcGetAllLibVersionsU
#define RfcGetDomainFingerprintHex RfcGetDomainFingerprintHexU
#define ItCreate		ItCreateU
#define ItDelete		ItDeleteU
#define ItGetLine		ItGetLineU
#define ItPutLine		ItPutLineU
#define ItCpyLine		ItCpyLineU
#define ItInsLine		ItInsLineU
#define ItAppLine		ItAppLineU
#define ItDelLine		ItDelLineU
#define ItGupLine		ItGupLineU
#define ItFree			ItFreeU
#define ItFill			ItFillU
#define ItLeng			ItLengU
#define ItRegTable		ItRegTableU
#define ItUserData		ItUserDataU
#define RfcConnect		RfcConnectU
#define RfcInstallTransactionControl RfcInstallTransactionControlU
#define RfcRaiseTables		RfcRaiseTablesU
#define RfcRaiseErrorMessage		RfcRaiseErrorMessageU
#define RfcConnArgv3		RfcConnArgv3U
#define RfcOpenExtV3		RfcOpenExtV3U
#define RfcGetAttributes	RfcGetAttributesU
#define RfcWaitForRequest	RfcWaitForRequestU
#define RfcInstallStructure	RfcInstallStructureU
#define RfcStructureToFile	RfcStructureToFileU
#define RfcStructureToFileByName RfcStructureToFileByNameU
#define RfcAcceptWithChannel	RfcAcceptWithChannelU
#define RfcDefineImportParam	RfcDefineImportParamU
#define RfcGetImportParam	RfcGetImportParamU
#define RfcSncMode		RfcSncModeU
#define RfcSncPartnerName	RfcSncPartnerNameU
#define RfcSncPartnerAclKey	RfcSncPartnerAclKeyU
#define RfcSncNameToAclKey	RfcSncNameToAclKeyU
#define RfcSncAclKeyToName	RfcSncAclKeyToNameU
#define RfcIoCtl		RfcIoCtlU
#define RfcSetCodePage		RfcSetCodePageU
#define RfcSetSystemCodePage	RfcSetSystemCodePageU
#define RfcGetDriverCallBack	RfcGetDriverCallBackU
#define RemObjInstallDriver	RemObjInstallDriverU
#define RemObjRegisterObject	RemObjRegisterObjectU
#define RemObjGetObject		RemObjGetObjectU
#define RemObjReleaseObject	RemObjReleaseObjectU
#define RfcCheckRegisterServer	RfcCheckRegisterServerU
#define RfcCancelRegisterServer RfcCancelRegisterServerU
#define RfcOpenEx		RfcOpenExU
#define RfcLastErrorEx		RfcLastErrorExU
#define RfcConvertBcdToChar	RfcConvertBcdToCharU
#define RfcConvertCharToBcd	RfcConvertCharToBcdU
#define RfcTxStart		RfcTxStartU
#define RfcTxPrepareToCommit	RfcTxPrepareToCommitU
#define RfcTxCommit		RfcTxCommitU
#define RfcTxAbort		RfcTxAbortU
#define RfcIndirectCallEx	RfcIndirectCallExU
#define RfcIndirectCallExExt	RfcIndirectCallExExtU
#define RfcUninstallFunction	RfcUninstallFunctionU
#define RfcGuiCheck		RfcGuiCheckU
#define RfcGuiPrepare		RfcGuiPrepareU
#define RfcGuiStart		RfcGuiStartU
#define RfcGuiReturn		RfcGuiReturnU
#define RfcGetNameEx		RfcGetNameExU
#define RfcResizeBlob		RfcResizeBlobU
#define RfcAllowStartProgram	RfcAllowStartProgramU
#define RfcSetTrace		RfcSetTraceU
#define RfcGetLongName		RfcGetLongNameU
#define RfcGetLongNameEx	RfcGetLongNameExU
#define RfcSapLangToIso		RfcSapLangToIsoU
#define RfcIsoLangToSap		RfcIsoLangToSapU
#define RfcCancel               RfcCancelU
#define RfcCharConversion       RfcCharConversionU
#define RfcRegisterPasswordChanger       RfcRegisterPasswordChangerU
#define RfcResizeXString	RfcResizeXStringU

#define RfcFreeString           RfcFreeStringU
#define RfcAllocString          RfcAllocStringU

#elif defined(SAPonOS390) && __STRING_CODE_SET__==ISO8859-1

#define RfcInit			RfcInitA
#define RfcEnvironment		RfcEnvironmentA
#define RfcLastError		RfcLastErrorA
#define RfcOpenExt		RfcOpenExtA
#define RfcOpen			RfcOpenA
#define RfcAccept		RfcAcceptA
#define RfcAcceptExt		RfcAcceptExtA
#define RfcClose		RfcCloseA
#define RfcAbort 		RfcAbortA
#define RfcCall			RfcCallA
#define RfcReceive		RfcReceiveA
#define RfcCallReceive		RfcCallReceiveA
#define RfcIndirectCall		RfcIndirectCallA
#define RfcGetName		RfcGetNameA
#define RfcGetData		RfcGetDataA
#define RfcDynamicGetData	RfcDynamicGetDataA
#define RfcSendData		RfcSendDataA
#define RfcRaise		RfcRaiseA
#define RfcListen		RfcListenA
#define RfcFileArgv		RfcFileArgvA
#define RfcConnArgv		RfcConnArgvA
#define RfcSendSystemInfo	RfcSendSystemInfoA
#define RfcInstallFunction	RfcInstallFunctionA
#define RfcInstallFunctionExt	RfcInstallFunctionExtA
#define RfcDispatch		RfcDispatchA
#define RfcGetTransID		RfcGetTransIDA
#define RfcCreateTransID	RfcCreateTransIDA
#define RfcCreateLocalTransID  RfcCreateLocalTransIDA
#define RfcGuidToTid    RfcGuidToTidA
#define RfcTidToGuid    RfcTidToGuidA
#define RfcCheckTransID		RfcCheckTransIDA
#define RfcConfirmTransID	RfcConfirmTransIDA
#define RfcAllocParamSpace	RfcAllocParamSpaceA
#define RfcFreeParamSpace	RfcFreeParamSpaceA
#define RfcAddExportParam	RfcAddExportParamA
#define RfcAddImportParam	RfcAddImportParamA
#define RfcAddTable		RfcAddTableA
#define RfcGetTableHandle	RfcGetTableHandleA
#define RfcCallExt		RfcCallExtA
#define RfcReceiveExt		RfcReceiveExtA
#define RfcCallReceiveExt	RfcCallReceiveExtA
#define RfcIndirectCallExt	RfcIndirectCallExtA
#define RfcGetDataExt		RfcGetDataExtA
#define RfcSendDataExt		RfcSendDataExtA
#define RfcExecProgram		RfcExecProgramA
#define RfcGetVersion		RfcGetVersionA
#define RfcCheckVersionsOfUsedLibs RfcCheckVersionsOfUsedLibsA
#define RfcGetAllLibVersions	RfcGetAllLibVersionsA
#define RfcGetDomainFingerprintHex RfcGetDomainFingerprintHexA
#define ItCreate		ItCreateA
#define ItDelete		ItDeleteA
#define ItGetLine		ItGetLineA
#define ItPutLine		ItPutLineA
#define ItCpyLine		ItCpyLineA
#define ItInsLine		ItInsLineA
#define ItAppLine		ItAppLineA
#define ItDelLine		ItDelLineA
#define ItGupLine		ItGupLineA
#define ItFree			ItFreeA
#define ItFill			ItFillA
#define ItLeng			ItLengA
#define ItRegTable		ItRegTableA
#define ItUserData		ItUserDataA
#define RfcConnect		RfcConnectA
#define RfcInstallTransactionControl RfcInstallTransactionControlA
#define RfcRaiseTables		RfcRaiseTablesA
#define RfcRaiseErrorMessage		RfcRaiseErrorMessageA
#define RfcConnArgv3		RfcConnArgv3A
#define RfcOpenExtV3		RfcOpenExtV3A
#define RfcGetAttributes	RfcGetAttributesA
#define RfcWaitForRequest	RfcWaitForRequestA
#define RfcInstallStructure	RfcInstallStructureA
#define RfcStructureToFile	RfcStructureToFileA
#define RfcStructureToFileByName RfcStructureToFileByNameA
#define RfcAcceptWithChannel	RfcAcceptWithChannelA
#define RfcDefineImportParam	RfcDefineImportParamA
#define RfcGetImportParam	RfcGetImportParamA
#define RfcSncMode		RfcSncModeA
#define RfcSncPartnerName	RfcSncPartnerNameA
#define RfcSncPartnerAclKey	RfcSncPartnerAclKeyA
#define RfcSncNameToAclKey	RfcSncNameToAclKeyA
#define RfcSncAclKeyToName	RfcSncAclKeyToNameA
#define RfcIoCtl		RfcIoCtlA
#define RfcSetCodePage		RfcSetCodePageA
#define RfcSetSystemCodePage    RfcSetSystemCodePageA
#define RfcGetDriverCallBack	RfcGetDriverCallBackA
#define RemObjInstallDriver	RemObjInstallDriverA
#define RemObjRegisterObject	RemObjRegisterObjectA
#define RemObjGetObject		RemObjGetObjectA
#define RemObjReleaseObject	RemObjReleaseObjectA
#define RfcCheckRegisterServer	RfcCheckRegisterServerA
#define RfcCancelRegisterServer RfcCancelRegisterServerA
#define RfcOpenEx		RfcOpenExA
#define RfcLastErrorEx		RfcLastErrorExA
#define RfcConvertBcdToChar	RfcConvertBcdToCharA
#define RfcConvertCharToBcd	RfcConvertCharToBcdA
#define RfcTxStart		RfcTxStartA
#define RfcTxPrepareToCommit	RfcTxPrepareToCommitA
#define RfcTxCommit		RfcTxCommitA
#define RfcTxAbort		RfcTxAbortA
#define RfcIndirectCallEx	RfcIndirectCallExA
#define RfcIndirectCallExExt	RfcIndirectCallExExtA
#define RfcUninstallFunction	RfcUninstallFunctionA
#define RfcGuiCheck		RfcGuiCheckA
#define RfcGuiPrepare		RfcGuiPrepareA
#define RfcGuiStart		RfcGuiStartA
#define RfcGuiReturn		RfcGuiReturnA
#define RfcGetNameEx		RfcGetNameExA
#define RfcResizeBlob		RfcResizeBlobA
#define RfcAllowStartProgram	RfcAllowStartProgramA
#define RfcSetTrace		RfcSetTraceA
#define RfcGetLongName		RfcGetLongNameA
#define RfcGetLongNameEx	RfcGetLongNameExA
#define RfcSapLangToIso		RfcSapLangToIsoA
#define RfcIsoLangToSap		RfcIsoLangToSapA
#define RfcCancel               RfcCancelA
#define RfcCharConversion       RfcCharConversionA
#define RfcRegisterPasswordChanger       RfcRegisterPasswordChangerA
#define RfcResizeXString	RfcResizeXStringA

#define RfcFreeString           RfcFreeStringA
#define RfcAllocString          RfcAllocStringA

#endif

#if defined(SAPwithUNICODE) && defined(SAPTYPEB_H)
/* SAPUNICODEOK */
/* type rfc_char_t | data type used for a single or multi byte
 * character field encoded with the default code page */
typedef SAP_UC rfc_char_t;
#else
/* SAPUNICODEOK */
/* @type rfc_char_t | data type used for a single or multi byte
 * character field encoded with the default code page */
typedef char rfc_char_t;
#endif

#ifndef sizeofR
#define sizeofR sizeof
#endif



/* @type rfc_unicode_t | data type used for unicode
 * character field encoded with 2 bytes per character (e. g. UTF16) */
typedef unsigned short rfc_unicode_t;

/* SAPUNICODEOK */
/* @type rfc_byte_t |
 * data type used for binary fields. */
typedef unsigned char rfc_byte_t;

/*
 * @enum RFCTYPE |
 *
 * RFC data types.
 */
typedef enum
{
  RFCTYPE_CHAR   = TYPC,         /* @emem 1-byte or multibyte character.
                                  * fixed sized, blank padded
                                  * ( <t RFC_CHAR> ).
			          */
  RFCTYPE_DATE   = TYPDATE,      /* @emem date ( YYYYYMMDD )
                                  * ( <t RFC_DATE> ).
				  */
  RFCTYPE_BCD    = TYPP,         /* @emem packed number.
				  * bcd number, any length between
				  * 1 and 16 bytes
                                  * ( <t RFC_BCD> ).
				  */
  RFCTYPE_TIME   = TYPTIME,      /* @emem time (HHMMSS)
                                  * ( <t RFC_TIME> ).
				  */
  RFCTYPE_BYTE   = TYPX   ,      /* @emem raw data, binary,
				  * fixed length, zero padded.
                                  * ( <t RFC_BYTE> ).
				  */
  RFCTYPE_ITAB   = TYPTABH,      /* @emem internal table
                                  * ( <t ITAB_H> ).
				  */
  RFCTYPE_NUM    = TYPNUM,       /* @emem digits, fixed size,
				  * '0' padded.
                                  * ( <t RFC_NUM> ).
				  */
  RFCTYPE_FLOAT  = TYPFLOAT,     /* @emem floating point,
				  * double precission
                                  * ( <t RFC_FLOAT> ).
				  */
  RFCTYPE_INT    = TYPINT,       /* @emem 4-byte integer
                                  * ( <t RFC_INT> ).
				  */
  RFCTYPE_INT2   = TYPINT2,      /* @emem 2-byte integer
				  * obsolete,
				  * not directly supported
				  * by ABAP/4.
				  */
  RFCTYPE_INT1   = TYPINT1,      /* @emem 1-byte integer
				  * obsolete,
				  * not directly supported
				  * by ABAP/4.
				  */
  RFCTYPE_DATE_1 = TYP1,         /* @emem old fashioned date,
				  * obsolete,
				  * not directly supported
				  * by ABAP/4.
				  */
  RFCTYPE_DATE_2 = TYP2,         /* @emem old fashioned date,
				  * obsolete,
				  * not directly supported
				  * by ABAP/4.
				  */

  /* starting from here ABAP types and RFC types differ */
  _RFCTYPE_differ_from_abap,

  RFCTYPE_NULL                   /* @emem not supported data type. */
  =   _RFCTYPE_differ_from_abap,

  RFCTYPE_WIDE_2,                /* wide character 2 byte (internal use)*/
  RFCTYPE_WIDE_4,                /* wide character 4 byte (internal use)*/
  RFCTYPE_STRUCTURE,             /* a structure (internal use) */

  RFCTYPE_ABAP4OBJECT,           /* abap/4 object (internal use) */

  RFCTYPE_IUNKNOWN,              /* @emem an IUnknown object
				  * can only be used in a COM/OLE
				  * environment.
                                  * ( <t RFC_IUNKNOWN> ).
			 	  */

  RFCTYPE_WSTRING,               /* @emem a wide character string
                                  * zero terminated, if given size
                                  * is big enough.
                                  * ( <t RFC_WSTRING> ).
 	                          */
  RFCTYPE_SAPAUTOMATION,         /* sap automation object
				  * sapgui only, internal use.
				  */
  RFCTYPE_STUB,			 /* remote object's stub
				  * (internal use)
				  */
  RFCTYPE_WIRE_OBJECT,           /* Object on the wire
                                  * (internal use)
                                  */
  RFCTYPE_BLOB,                  /* @emem binary large object,
				  * address must point to a <t RFC_BLOB>
				  * structure.
				  */
  RFCTYPE_CORBA_OBJECT,          /* a Corba object
				  * (internal use only).
				  */
  RFCTYPE_PADDING,               /* @emem use in dummy fields to ensure proper
                                  * alignment of structure with sub structures
 	                          */
  RFCTYPE_UNICODE,               /* @emem 2-bytes characters for strings
                                  * in unicode environment.
                                  * Blank padded fixed size (e. g. UTF16)
                                  * ( <t RFC_UNICODE> ).
			          */

  RFCTYPE_XMLDATA,      /* @emem zero terminated string describing ABAP data
                         * types/structure as XML */

  RFCTYPE_STRING,        /* @emem zero terminated string */

  RFCTYPE_XSTRING,      /* @emem raw string with length field in bytes */


  _RFCTYPE_max_value,     /* the max. value of RFCTYPEs */

  /* some special values */
#if defined(SAPonRM600) || defined (SAPonPTX)
                 		/* cast required incl. CDS++ 1.0B C1016615 */
   RFCTYPE_WCHAR                  /* @emem wide character string,
                                   * blank padded
                                   * ( <t RFC_WCHAR> ).
                                   */
    = ( sizeofR( wchar_t ) == 2 ) ? (int)RFCTYPE_WIDE_2 : (int)RFCTYPE_WIDE_4,
   _RFCTYPE_illegal_wchar         /* not allowed wide character type */
    = ( sizeofR( wchar_t ) == 2 ) ? (int)RFCTYPE_WIDE_4 : (int)RFCTYPE_WIDE_2,
#else
  RFCTYPE_WCHAR                  /* @emem wide character string,
				  * blank padded
                                  * ( <t RFC_WCHAR> ).
				  */
  = ( sizeofR( wchar_t ) == 2 ) ? RFCTYPE_WIDE_2 : RFCTYPE_WIDE_4,

  _RFCTYPE_illegal_wchar         /* not allowed wide character type */
  = ( sizeofR( wchar_t ) == 2 ) ? RFCTYPE_WIDE_4 : RFCTYPE_WIDE_2,
#endif


  _RFCTYPE_the_last_line_without_colon

}
RFCTYPE;

/*
 * @enum RFC_BLOB_CLASS |
 *
 * Memory class of a BLOB.
 */
typedef enum
{
    RFC_BLOB_CLASS_ANY    = 0,  /* @emem any memory class (send only) */
    RFC_BLOB_CLASS_MALLOC = 9   /* @emem object is allocated in memory
                                 * using 'malloc' or 'realloc'. Memory has
                                 * to be released by 'free'.
                                 */
}
RFC_BLOB_CLASS;

/* @type RFC_BLOB |
 * a structure describing a BLOB (binary large object)
 */
typedef struct
{
    RFC_BLOB_CLASS blobClass;  /* @field type of blob (must be
				* <e RFC_BLOB_CLASS.RFC_BLOB_CLASS_MALLOC>)
				*/
    RFCTYPE        blobType;   /* @field RFC type of blob object,
                                * must be either
                                * <e RFCTYPE.RFCTYPE_CHAR> or
                                * <e RFCTYPE.RFCTYPE_BYTE>.
                                */
    unsigned long  blobSize;       /* @field size in bytes */
    void         * blobAddress;    /* @field address of buffer */
}
RFC_BLOB;

/* @type RFC_TYPEHANDLE |
 * a value describing an ABAP/4 field. It may describe
 * a scalar field like <t RFC_CHAR> (s. enum <t RFCTYPE> )
 * or a structure. A
 * structure description must be installed by the
 * function <f RfcInstallStructure>.
 *
 * @comm
 * In former versions corresponding field where simply
 * defined as 'unsigned'. Use <t RFC_TYPEHANDLE> instead.
 */
typedef unsigned RFC_TYPEHANDLE;

/* @type RFC_TYPE_ELEMENT |
 * an element of a strucure used in an RFC interface.
 * This type is used to describe a structure, which is
 * to be installed by the
 * function <f RfcInstallStructure>.
 */
typedef struct
{
    rfc_char_t *   name;     /* @field name of the field
                              * (only used for trace).
                              */
    RFC_TYPEHANDLE type;     /* @field (scalar) data type of the element. */
    unsigned       length;   /* @field length of the field in bytes. */
    unsigned       decimals; /* @field decimals
                              * (only bcd fields, TYPP ).
                              */

    unsigned _not_used;
}
RFC_TYPE_ELEMENT;


/* @type RFC_TYPE_ELEMENT2 |
 * an element of a strucure used in an RFC interface.
 * This type is used to describe a structure, which is
 * to be installed by the
 * function <f RfcInstallStructure2>.
 */
typedef struct
{
    rfc_char_t *   name;     /* @field name of the field
                              * (only used for trace).
                              */
    RFC_TYPEHANDLE type;     /* @field (scalar) data type of the element. */
    unsigned       length;   /* @field length of the field in bytes. */
    unsigned       decimals; /* @field decimals
                              * (only bcd fields, TYPP ).
                              */

    unsigned       offset;   /* @field byte offset of the field from */
                             /* the beginning of the structure */ 	
}
RFC_TYPE_ELEMENT2;

/* RFC scalar data types */

/* @type RFC_CHAR |
 * ABAP/4 data type C,
 * blank padded character string of fixed length.
 * Used for data transported by RFC
 *
 * characters are encoded in UTF16 in unicode builds
 *
 * characters are encoded in the system default code page
 * or in the code page selected by <f RfcSetCodePage>, <f RfcSetSystemCodePage>,
 * <f RfcOpenEx> or defined by environment variable SAP_CODEPAGE
 * in multibyte builds. The default codepage is 1100.
 */
/* @type RFC_UNICODE |
 * ABAP/4 data type C,
 * blank padded character string of fixed length in unicode environment.
 */
/* @type RFC_NUM |
 * ABAP/4 data type N,
 * character string of fixed length containing only digits.
 */
/* @type RFC_BYTE |
 * ABAP/4 data type X,
 * raw data.
 */
/* @type RFC_BCD |
 * ABAP/4 data type P,
 * packed number in BCD format (binary coded decimal).
 */
/* @type RFC_INT1 |
 * Data dictionary data type INT1,
 * 1-byte unsigned integer.
 */
/* @type RFC_INT2 |
 * Data dictionary data type INT2,
 * 2-byte signed integer.
 */
/* @type RFC_INT |
 * ABAP/4 data type I,
 * 4-byte signed integer.
 */
/* @type RFC_FLOAT |
 * ABAP/4 data type F,
 * 8-byte IEEE floating point number.
 */
/* @type RFC_DATE |
 * ABAP/4 data type D,
 * 8-byte date (YYYYMMDD).
 */
/* @type RFC_TIME |
 * ABAP/4 data type T,
 * 6-byte time (HHMMSS).
 */
/* @type RFC_WCHAR |
 * ABAP/4 data type W,
 * double-byte, unicode string,
 * blank padded.
 */
/* @type RFC_WSTRING |
 * similiar than ABAP/4 data type W.
 *
 * Double-byte, unicode string,
 * zero terminated
 * (if the given length
 * in <t RFC_PARAMETER>.
 * is big enough )
 *
 * Can be used instead of <t RFC_WCHAR>, if more appropriate.
 */
/* @type RFC_STRING |
 * similar than ABAP/4 data type string with variable length.
 * The difference is RFC_STRING is a zero terminated string and ABAP/4 string
 * is a blank padded string. For better understanding you can consider RFC_STRING
 * like  typedef unsigned char* RFC_STRING.
 * The conversion between zero terminated
 * (extern RFC program) and blank padded (ABAP) is done automatically.
 * All strings are to be handled by sending and receiving in external programs
 * as UTF8 zero terminated strings.
 *
 */
/* @type RFC_XSTRING |
 * raw string with length field in bytes.
 */
/* @type RFC_IUNKNOWN |
 * COM object.
 * Can be mapped to ABAP/4 objects and vice versa.
 * (Available, only if COM adapter is installed)
 */

#ifdef RFC_CHAR_UNSIGNED
typedef unsigned char  RFC_CHAR;     /* characters, TYPC        */
#else
typedef rfc_char_t     RFC_CHAR;     /* characters, TYPC        */
#endif
typedef rfc_unicode_t  RFC_UNICODE;  /* characters, TYPC in unicode env. */
typedef RFC_CHAR       RFC_NUM;      /* digits, TYPNUM          */
typedef rfc_byte_t     RFC_BYTE;     /* raw data, TYPX          */
typedef rfc_byte_t     RFC_BCD;      /* packed numbers, TYPP    */
typedef rfc_byte_t     RFC_INT1;     /* 1 byte integer, TYPINT1 */
typedef short          RFC_INT2;     /* 2 byte integer, TYPINT2 */
typedef double         RFC_FLOAT;    /* floating point, TYPFLOAT*/
typedef RFC_CHAR       RFC_DATE[8];  /* date, TYPDATE (YYYYMMDD)*/
typedef RFC_CHAR       RFC_TIME[6];  /* time, TYPTIME (HHMMSS)  */
typedef wchar_t        RFC_WCHAR;    /* wchar_t                 */
typedef wchar_t *      RFC_WSTRING;  /* wchar_t, zero term.     */
typedef char *         RFC_XMLDATA;  /* ABAP structs encoded as XML Document */
typedef rfc_char_t*    RFC_STRING;   /* zero terminated string in UTF8-Format */

#ifdef _OBJBASE_H_
/* only if com header is included */
typedef IUnknown *     RFC_IUNKNOWN; /* IUnknown, COM object    */
#endif

#ifdef __midl
 typedef long int      RFC_INT;
#else

#define RFC_INT_MAX 0x7fffffff
#if INT_MAX == RFC_INT_MAX
 typedef int           RFC_INT;     /* 4 byte integers, TYPINT  */
#elif LONG_MAX == RFC_INT_MAX
 typedef long int      RFC_INT;     /* 4 byte integers, TYPINT  */
#else
#error local integer representation is not supported by RFC API
#endif
#endif /* __midl or C */

typedef struct
{
 RFC_INT            length;
 RFC_BYTE          *content;
} RFC_XSTRING; /* binary string */




/* saptype data types */
#ifndef SAPTYPE_H
 typedef RFC_CHAR       SAP_CHAR;
 typedef RFC_BYTE       SAP_BYTE;
 typedef RFC_BCD        SAP_BCD;
 typedef RFC_INT        SAP_INT;
 typedef RFC_INT2       SAP_SHORT;
 typedef RFC_FLOAT      SAP_DOUBLE;
 typedef RFC_DATE       SAP_DATE;
 typedef RFC_TIME       SAP_TIME;
 typedef RFC_BYTE       SAP_RAW;

#ifdef __midl
 typedef unsigned long int   SAP_UINT;
 typedef unsigned short int  SAP_USHORT;
#else
 typedef unsigned short SAP_USHORT;
#if INT_MAX == RFC_INT_MAX
 typedef unsigned int  SAP_UINT;
#elif LONG_MAX == RFC_INT_MAX
 typedef unsigned long SAP_UINT;
#else
#error local integer representation is not supported by RFC API
#endif
#endif /* __midl or C */

 #define SAP_UUID_TYPEDEFD
 typedef struct _SAP_UUID
 {
    SAP_UINT   a;
    SAP_USHORT b;
    SAP_USHORT c;
    SAP_BYTE   d[8];
 }
 SAP_UUID;


#endif

/* Initialization */
void SAP_API RfcInit( void );

/* Environment : memory allocation and error handling */

typedef void* (DLL_CALL_BACK_FUNCTION_PTR RFC_ALLOCATE)
                      ( void * old_ptr,
                        unsigned long size );
typedef void  (DLL_CALL_BACK_FUNCTION_PTR RFC_ERROR_HANDLER)
                      ( rfc_char_t * error_id );

typedef struct
{
   RFC_ERROR_HANDLER errorhandler;
   RFC_ALLOCATE      allocate;
}
RFC_ENV;

void SAP_API RfcEnvironment( RFC_ENV * new_env );


 /* @struct RFC_ERROR_INFO |
 *
 *
 * !!! Please use <t RFC_ERROR_INFO_EX> and <f RfcLastErrorEx>
 *     instead of this structure.
 *
 *
 * structure returned by <f RfcLastError> describing
 * the last RFC error that occured.
 */
typedef struct
{
    rfc_char_t key[33];       /* @field error code to identify
                               * the error */
    rfc_char_t status[128];   /* @field state of the RFC
                               * connection
                               */
    rfc_char_t message[256];  /* @field text describing the error */
    rfc_char_t intstat[128];  /* @field internal description of the
                               * RFC connection
                               */
}
RFC_ERROR_INFO;

/*------------------------------------------------------
 * @func
 *
 *
 * !!! Please use <f RfcLastErrorEx> instead of this function.
 *
 *
 * RfcLastError describes the last error reported by
 * some function of the RFC API.
 *
 * @rdesc
 * returns 1 if no error occurred and 0 elsewhere.
 *
 * @xref <f RfcLastErrorEx>
 *
 *------------------------------------------------------*/
int SAP_API RfcLastError(
  RFC_ERROR_INFO * error_info  /* @parm structure
                                * <t RFC_ERROR_INFO> describing the
                                * error.
                                */
                        );


/* @enum RFC_ERROR_GROUP | RFC error group */
typedef enum
{
   RFC_ERROR_PROGRAM               = 101, /* @emem Error in RFC program
                                           */
   RFC_ERROR_COMMUNICATION         = 102, /* @emem Error in Network & Communications
                                           */
   RFC_ERROR_LOGON_FAILURE         = 103, /* @emem SAP logon error
                                           */
   RFC_ERROR_SYSTEM_FAILURE        = 104, /* @emem e.g. SAP system exception raised
                                           */
   RFC_ERROR_APPLICATION_EXCEPTION = 105, /* @emem The called function module raised
                                           * an exception
                                           */
   RFC_ERROR_RESOURCE              = 106, /* @emem Resource not available
                                           * (e.g. memory insufficient,...)
                                           */
   RFC_ERROR_PROTOCOL              = 107, /* @emem RFC Protocol error
                                           */
   RFC_ERROR_INTERNAL              = 108, /* @emem RFC Internal error
                                           */
   RFC_ERROR_CANCELLED             = 109, /* @emem RFC Registered Server was
                                           * cancelled
                                           */
   RFC_ERROR_BUSY                  = 110  /* @emem System is busy, try later
                                           */
}
RFC_ERROR_GROUP;

 /* @struct RFC_ERROR_INFO_EX |
 *
 * structure returned by <f RfcLastErrorEx> describing
 * the last RFC error.
 */
typedef struct
{
    RFC_ERROR_GROUP  group;         /* @field error group <t RFC_ERROR_GROUP> */
    rfc_char_t       key[33];       /* @field error code to identify the error.
				     * The error group (type integer) and error key
				     * (type string) have the same meaning.
				     * RFC program can better analyze the error
				     * with the error group instead of error key.
				     */
    rfc_char_t       message[513];  /* @field error text to describe the error */
}
RFC_ERROR_INFO_EX;


/*------------------------------------------------------
 * @func
 *
 * RfcLastErrorEx describes the last error reported by
 * some function of the RFC API.
 *
 * @rdesc
 * 1 if no error occurred or there is no error info available or 
 * this api was already called for the same error.
 * 
 * 0 elsewhere.
 *
 * @comm 
 * Do not call this api after call contains
 * <t RFC_ERROR_INFO_EX> structure in the parameter list.
 *
 * In this case 
 *
 * - RfcLastErrorEx returns 1 although previous call failed 
 *
 * - please retrieve the error information from the parameter.
 *
 * An example for api contains <t RFC_ERROR_INFO_EX> in the
 * parameter list is <f RfcOpenEx>.
 *------------------------------------------------------
 */
int SAP_API RfcLastErrorEx
(
 RFC_ERROR_INFO_EX * error_info  /* @parm structure
                                   * <t RFC_ERROR_INFO_EX> describing the
                                   * error.
                                   */
);



/* @type RFC_HANDLE | handle for RFC connection
 * @flag RFC_HANDLE_NULL | the RFC handle is not set
 */
typedef unsigned RFC_HANDLE;

/* Null-Value : handle not set */
#define RFC_HANDLE_NULL   0

 /* @struct RFC_ATTRIBUTES |
  * structure returned by <f RfcGetAttributes> describing
  * some information about this RFC connection.
  */
typedef struct
{
  rfc_char_t dest[64+1];            /* @field RFC destination              */
  rfc_char_t own_host[100+1];       /* @field Own host name                */
  rfc_char_t partner_host[100+1];   /* @field Partner host name            */
  rfc_char_t systnr[2+1];           /* @field R/3 system number            */
  rfc_char_t sysid[8+1];            /* @field R/3 system name              */
  rfc_char_t client[3+1];           /* @field Client                       */
  rfc_char_t user[12+1];            /* @field User                         */
  rfc_char_t language[1+1];         /* @field Language                     */
  rfc_char_t trace;                 /* @field ON/OFF: 'X'/' '              */
  rfc_char_t ISO_language[2+1];     /* @field 2-byte ISO-Language          */
  rfc_char_t own_codepage[4+1];     /* @field Own code page                */
  rfc_char_t partner_codepage[4+1]; /* @field Partner code page            */
  rfc_char_t rfc_role;              /* @field C/S: RFC Client / RFC Server */
  rfc_char_t own_type;              /* @field 2/3/E/R: R/2,R/3,Ext,Reg.Ext */
  rfc_char_t own_rel[4+1];          /* @field My system release            */
  rfc_char_t partner_type;          /* @field 2/3/E/R: R/2,R/3,Ext,Reg.Ext */
  rfc_char_t partner_rel[4+1];      /* @field Partner system release       */
  rfc_char_t kernel_rel[4+1];       /* @field Partner kernel release       */
  rfc_char_t CPIC_convid[9];        /* @field CPI-C Conversation ID        */
  rfc_char_t password_sate;         /* @field state from the password      */
                                    /*        see RFC_PASS_*               */
  rfc_char_t reserved[214];         /* @field reserve                      */
}
RFC_ATTRIBUTES;

/*------------------------------------------------------
 * @func
 *
 * RfcGetAttributes returns some information about this
 * RFC connection.
 *
 * The RFC library can only set these attributes after the
 * RFC connection is established and at least one RFC function
 * is called.
 *
 * Therefore, it should be called after <f RfcOpenEx> with
 * SAP-logon check at OPEN-time or after the first <f RfcReceive>
 * in an RFC client program or after <f RfcGetData> in an
 * RFC server program.
 *
 *
 * @rdesc
 * returns 0 if no error occurred and 1 elsewhere.
 *------------------------------------------------------
 */
int SAP_API RfcGetAttributes(
  RFC_HANDLE       handle,         /* @parm connection handle. */
  RFC_ATTRIBUTES * rfc_attributes  /* @parm structure
                                * <t RFC_ATTRIBUTES> describing some
                                * information about this RFC connection.
                                */
                        );

/*
 * @enum RFC_MODE |
 *
 * Type of connection to be opened by <f RfcOpen>.
 */
typedef enum
{
  RFC_MODE_R3ONLY = 0,      /* @emem Use R/3 protocol and addressing scheme.
                             * Only for R/3 systems. Any kind
                             * of user id (dialog user id, cpic user id)
                             * is possible.
                             *
                             * <e RFC_OPTIONS.connopt> must point to a structure
                             * of type <t RFC_CONNOPT_R3ONLY>.
                             *
                             */
  RFC_MODE_CPIC = 1,        /* @emem Use R/2 protocol and addressing scheme.
                             * Must be used for R/2,
                             * cpic user id are only allowed.
                             * (Since an R/3 system is also capable of
                             * understanding the R/2 RFC protocol, you can
                             * reach also an R/3 system with that mode,
                             * you must use a 'sideinfo' file for addressing,
                             * however.
                             *
                             * <e RFC_OPTIONS.connopt> must point to a structure
                             * of type <t RFC_CONNOPT_CPIC>.
                             */
  RFC_MODE_VERSION_3 = 3,   /* @emem Use R/3 protocoll Version 3.
                             * The receiving SAP system must have at least
                             * Rel. 3.0C to be able to serve every kind
                             * of options.
                             *
                             * <e RFC_OPTIONS.connopt> must point to a structure
                             * of type <t RFC_CONNOPT_VERSION_3>.
                             */
  RFC_MODE_PARAMETER = 4,   /* @emem Use R/3 protocoll Version 3 or R/2 protocol and
                             * addressing scheme.
                             * This mode includes all other modes above and all necessary
                             * parameters will be read from a INI-file (saprfc.ini).
                             * See saprfc.ini for more details
                             *
                             * In some cases the receiving SAP system must have at least
                             * Rel. 3.0C to be able to serve every kind of options.
                             *
                             * <e RFC_OPTIONS.destination> must point to a valid entry
                             * in the saprfc.ini. This file can be in the current directory
                             * where RFC programs are running or defined by the Environment
                             * Variable RFC_INI (e.g. /usr/rfctest/saprfc.ini).
                             *
                             * <e RFC_OPTIONS.connopt> must be set to NULL.
                             */
  RFC_MODE_CPIC_EXT = 5     /* @emem Use R/2 protocol and addressing scheme
                             * for communication between two external programs
                             * without using the 'sideinfo' or the 'saprfc.ini' file.
                             *
                             * <e RFC_OPTIONS.connopt> must point to a structure
                             * of type <t RFC_CONNOPT_CPIC_EXT>.
                             */
}
RFC_MODE;


/* internal use */
#define RFC_OPEN_USE_RFC       'R'
#define RFC_OPEN_USE_CPIC      ' '

/*
 * @enum RFC_SNC_MODE |
 *
 * SNC mode of an RFC connection: ON or OFF
 *
 * The default value is RFC_SNC_MODE_OFF.
 *
 */
typedef enum
{
  RFC_SNC_MODE_OFF = 0,     /* @emem Working without SNC. */

  RFC_SNC_MODE_ON  = 1      /* @emem Working with SNC.    */
}
RFC_SNC_MODE;

/*
 * @enum RFC_SNC_QOP |
 *
 * SNC Quality of service (s. SNC document for more details)
 *
 * The default value is RFC_SNC_QOP_DEFAULT.
 *
 */
typedef enum
{
  RFC_SNC_QOP_INVALID   = 0,    /* @emem plain text, but with authorisation */
  RFC_SNC_QOP_OPEN      = 1,    /* @emem plain text, but with authorisation */
  RFC_SNC_QOP_INTEGRITY = 2,    /* @emem each packet is integrity protected */
  RFC_SNC_QOP_PRIVACY   = 3,    /* @emem each packet is privacy protected   */

  RFC_SNC_QOP_DEFAULT   = 8,    /* @emem use global configuration defaults  */
  RFC_SNC_QOP_MAX       = 9     /* @emem use maximum available security     */
}
RFC_SNC_QOP;


/* @struct RFC_EXT_DATA |
 * includes extended parameters for opening an RFC connection
 * such as SNC parameters (SNC: Secure Network Communications)
 *
 * If working with SNC and not using the 'saprfc.ini' file,
 * all SNC parameters must be set in this structure.
 */
typedef struct
{
  RFC_SNC_MODE   snc_mode;            /* @field SNC mode for working
                                       * with or without SNC.
                                       */
  RFC_SNC_QOP    snc_qop;             /* @field SNC Quality of service
                                       */
  rfc_char_t         * snc_myname;          /* @field Own SNC name if you don't want
				       * to use the default SNC name
                                       */
  rfc_char_t         * snc_partnername;     /* @field SNC name of partner system
                                       *
                                       */
  rfc_char_t         * snc_lib;             /* @field name of the SNC library
                                       * on the local system
                                       */
  rfc_char_t          xmlflag[2];

  rfc_char_t          reserve[254];        /* @field reserve, must be set to 0
                                       *
                                       */
}
RFC_EXT_DATA;


/* @struct RFC_CONNOPT_CPIC |
 * Options for an R/2 connection (via SAP gateway).
 * connection data must be specified at the SAP gateway,
 * 'destination' is used as key into the 'sideinfo' file there.
 */
typedef struct
{
   rfc_char_t *   gateway_host;       /* @field gateway hostname
                                 */
   rfc_char_t *   gateway_service;    /* @field gateway service
                                 * (in general sapgw00).
                                 */
}
RFC_CONNOPT_CPIC;


/* @struct RFC_CONNOPT_CPIC_EXT |
 * Options for an RFC connection (via SAP gateway)
 * between two external programs without using the 'sideinfo'
 * or the 'saprfc.ini' file.
 */
typedef struct
{
   int      register_mode;      /* @field register mode:
                                 *  0: Server program is to be started by SAP gateway.
                                 *  1: Server program is running in register mode.
                                 */
   rfc_char_t *   gateway_host;       /* @field gateway hostname
                                 */
   rfc_char_t *   gateway_service;    /* @field gateway service
                                 */
   rfc_char_t *   tp_host;            /* @field host name of the partner program
                                 * not needed if register_mode is 1.
                                 */
   rfc_char_t *   tp_name;            /* @field name or program ID of the partner program
                                 * if the parameter 'register_mode' is set to 1.
                                 */
   void * ext_data;             /* @field data for working with extended features
                                 * such as SNC (Secure Network Communications).
                                 * 'ext_data' must be set to NULL or point to a
                                 * structure of type <t RFC_EXT_DATA>.
                                 */
}
RFC_CONNOPT_CPIC_EXT;


/* @struct RFC_CONNOPT_R3ONLY |
 * Options for a connection to an R/3 system.
 */
typedef struct
{
   rfc_char_t *   hostname;           /* @field Host name of target system.
                                 * Host names can be regular host
                                 * names
                                 * defined in a 'hosts' file,
                                 * an IP address like
                                 * 123.123.123.123 or a saprouter
                                 * address as
                                 * /H/hostname/S/port/H/host/S/port/...
                                 */
   int      sysnr;              /* @field system number (0-99).
                                 * This is the number by which
                                 * the target system
                                 * is identified. In general this is 0.
                                 */
   rfc_char_t *   gateway_host;       /* @field gateway hostname.
                                 * If the pointer is NULL, the gateway is
                                 * assumed to run at 'hostname'.
                                 */
   rfc_char_t *   gateway_service;    /* @field gateway service.
                                 * If the pointer is NULL,
                                 * the service "sapgw##" with
                                 * ## = 'sysnr' is assumed.
                                 */
}
RFC_CONNOPT_R3ONLY;


/* @struct RFC_CONNOPT_VERSION_3 |
 * Options for a connection to a R/3 system.
 * The target system must be of release 3.0C or later.
 *
 * Since a R/3 system of version 3.0 always has it's
 * own 'gateway' process, no gateway options are necessary
 * any more.
 */
typedef struct
{
   rfc_char_t *   hostname;           /* @field Host name of target system.
                                 * Host names can be regular host
                                 * names
                                 * defined in a 'hosts' file,
                                 * an IP address like
                                 * 123.123.123.123 or a saprouter
                                 * address as
                                 * /H/hostname/S/port/H/host/S/port/...
                                 *
                                 * If 'use_load_balancing' is set to a non-zero
                                 * value, this field can be NULL.
                                 *
                                 */
   RFC_INT  sysnr;              /* @field system number (0-99).
                                 * This is the number by which
                                 * the target system
                                 * is identified. In general this is 0.
                                 *
                                 * If 'use_load_balancing' is set to a non-zero
                                 * value, this field is ignored.
                                 *
                                 */
   RFC_INT use_load_balancing;  /* @field Use the load balancing features
                                 * of a SAP system.
                                 *
                                 * If you set this to a non-zero value,
                                 * 'hostname' and 'sysnr' are ignored.
                                 *
                                 * You must set the fields 'lb_host'
                                 * and 'lb_system_name' instead.
                                 *
                                 * The target system is determined dynamically
                                 * then.
                                 *
                                 */
   rfc_char_t *  lb_host;             /* @field Host name where the 'message server'
                                 * of the target system is running.
                                 *
                                 * This field must only be filled, if
                                 * 'use_load_balancing' is set to a non-zero
                                 * value.
                                 */
   rfc_char_t *  lb_system_name;      /* @field Name of the target system (C11 i.e.)
                                 *
                                 * This field must only be filled, if
                                 * 'use_load_balancing' is set to a non-zero
                                 * value.
                                 */
   rfc_char_t *  lb_group;            /* @field application server group.
                                 *
                                 * The group of application servers to be used.
                                 *
                                 * This field must only be filled, if
                                 * 'use_load_balancing' is set to a non-zero
                                 * value.
                                 */
   RFC_INT use_sapgui;          /* @field use 'sapgui' process to
                                 * display SAP dynpros and graphics.
                                 *
                                 * Set this to 1 to activate
                                 * this functionality.
                                 *
                                 * Set this to 2 if you want to let SAPGUI
                                 * automatically invisible after each RFC function.
                                 * This functionality requires R/3, Rel. >= 3.0F
                                 * and only available with SAPGUI, Rel. >= 3.0E
                                 * on Windows NT or 95.
                                 *
                                 * Starting sapgui takes some time, so
                                 * you should not set that value, if you do not
                                 * need it.
                                 *
                                 * This field is set implicitly as soon as
                                 * you activate the ABAP/4 debugger by
                                 * entering 'D' in the <e RFC_OPTIONS.trace> field or
                                 * by setting RFC_DEBUG in the system environment.
                                 *
                                 */
    void * ext_data;            /* @field data for working with extended features
                                 * such as SNC (Secure Network Communications).
                                 * 'ext_data' must be set to NULL or point to a
                                 * structure of type <t RFC_EXT_DATA>.
                                 */
}
RFC_CONNOPT_VERSION_3;


/* @struct RFC_OPTIONS |
 * parameters for <f RfcOpen>.
 *
 * Depending on the type of connection various data have to be supplied
 * to open a RFC connection.
 *
 * There are 3 ways to supply this information:
 *
 * 1. You can enter a destination name pointed to an entry in a 'saprfc.ini'
 *    file which contains the necessary network parameters and RFC specific
 *    parameters for opening the connection at <f RfcOpen>.
 *
 * 2. You can enter a destination name pointed to an entry in a 'sideinfo'
 *    file which only contains the necessary network parameters for opening
 *    the connection at <f RfcOpen>.
 *
 * 3. You supply in your program all the data needed for opening the
 *    connection at <f RfcOpen>.
 *
 *
 *
 * We recommend the first one (working with saprfc.ini) because you can use some
 * RFC features today and in the future without changes in your RFC programs.
 *
 * See saprfc.ini for more details.
 *
 *
 *
 * Fields that are not supplied must be set to NULL.
 *
 * The structure consists of 3 groups.
 *
 * The data needed to establish the physical connection depend on
 * the type of connection (R/2 or R/3 connection, version , ... ).
 * Depending on the contents of the field 'mode', the field 'connopt'
 * must point to different structures.
 *
 * The sign on data ( client, user, password, language ) are needed
 * for authentication. Since <f RfcOpen> only opens the physical connection
 * directly. these data are only checked if the first RFC call is
 * send.
 *
 * The third field contains a trace flag. If not zero, a trace file
 * is written for all the operations corresponding to this connection
 * to ease error analysis.
 *
 * Please make sure, that this structure is completely initialized
 * with 0 to allow adding more connection parameters in the future
 * by SAP.
 *
 * @ex Options for a R/3 connection. |
 *
 * // static = initialized structures
 * static RFC_OPTIONS        options;
 * static RFC_CONNOPT_R3ONLY rfc_connopt_r3only;
 * RFC_HANDLE                handle;
 *
 * options.destination = "TEST";
 * options.mode        = RFC_MODE_R3ONLY;
 * options.client      = "000";
 * options.user        = "SAP*";
 * options.language    = "E";
 * options.password    = "PASS";
 * options.trace       = 0;                    // turn trace off
 *
 * options.connopt     = &rfc_connopt_r3only;
 *
 * rfc_connopt_r3only.hostname = "some_host";  // some host name
 * rfc_connopt_r3only.sysnr    = 0;            // system 00
 *
 * handle = RfcOpen( &options );
 * if ( handle == RFC_HANDLE_NULL )
 * {
 *    .....
 *
 */
typedef struct
{
   rfc_char_t *   destination;        /* @field name of destination.
                                 * If the connection is not described
                                 * completely, this name is used as a
                                 * key for a 'sideinfo', where the
                                 * connection should be described then.
                                 * You always have to fill this field.
                                 */

   /* connection data */
   RFC_MODE mode;               /* @field connection mode.
                                 * There are two different protocol
                                 * types for RFC, depending on whether
                                 * your target system is a R/2 or a R/3
                                 * or an external system.
                                 *
                                 * If your target system is a SAP system
                                 * of Release 3.0C or later you can use
                                 * variuos special options, if you enter
                                 * the value <e RFC_MODE.RFC_MODE_VERSION_3>
                                 * here.
                                 *
                                 * Depending on the content of
                                 * this field, connopt must point to
                                 * different structures.
                                 * (s. <t RFC_MODE>).
                                 */
   void *   connopt;            /* @field If connopt is NULL, the 'sideinfo'
                                 * or the 'saprfc.ini' file is used to determine the
                                 * the connection parameters.
                                 *
                                 * Without 'sideinfo' file
                                 * 'connopt' must point to a structure
                                 * of type <t RFC_CONNOPT_R3ONLY>,
                                 * <t RFC_CONNOPT_VERSION_3> or
                                 * <t RFC_CONNOPT_CPIC_EXT> or
                                 * <t RFC_CONNOPT_CPIC> depending
                                 * on the value of 'mode'.
                                 *
                                 */

   /* sign on data */
   rfc_char_t *   client;             /* @field signon data: client   */
   rfc_char_t *   user;               /* @field signon data: user id  */
   rfc_char_t *   password;           /* @field signon data: password */
   rfc_char_t *   language;           /* @field signon data: language */

   /* options */
   int      trace;              /* @field trace.
                                 * If 0, no trace is written,
                                 * if not 0, the RFC library traces
                                 * all activities into a file 'dev_rfc'
                                 * in the actual working directory.
                                 *
                                 * If your target system is of release 3.0C
                                 * or later, you can enter the value 'D' here
                                 * to start the ABAP/4 debugger on the target
                                 * system.
                                 *
                                 * The ABAP/4 debugger can also be activated
                                 * by setting the environment variable
                                 * RFC_DEBUG before the call to <f RfcOpen>
                                 * is done.
                                 */
}
RFC_OPTIONS;


#define RFC_HOSTNAME_LN        100
typedef rfc_char_t RFC_HOSTNAME[RFC_HOSTNAME_LN+1];


/* @func
 *
 *
 * =====================================================
 *
 *   Please use <f RfcOpenEx> instead of this function
 *
 * =====================================================
 *
 *
 * opens a RFC connection via LOAD BALANCING
 *
 * (only available for R/3 Release 3.0 onwards).
 *
 * With this function the RFC library will try to connect to an
 * application server with the least load (LOAD BALANCING principle)
 * within a group of predefined application servers.
 *
 * This function has following advantages:
 *
 * 1. The load in R/3 system is distributed to different application server
 *
 * 2. RFC connections are thus independent of a specific application server.
 *    (With <f RfcOpen> or <f RfcOpenExt> a RFC connection could only be established
 *    to a predefined application server).
 *
 * 3. Only the host name of the message server and its port number of the
 *    according R/3 are required in the host file and the services file.
 *
 *    Information about SAP gateway, application server, system number,... as
 *    parameter for <f RfcOpen>, <f RfcOpenExt> or as parameter in sideinfo are no longer
 *    necessary. Even the sideinfo is no longer required.
 *
 *
 * @rdesc
 * returns a handle to the RFC connection (<t RFC_HANDLE>) or
 * RFC_HANDLE_NULL, if the connection cannot be opened.
 *
 * @xref <f RfcOpenEx>
 */
RFC_HANDLE SAP_API RfcConnect(
rfc_char_t *       system_name, /* @parm name of the R/3 system */
rfc_char_t *       ms_hostname, /* @parm hostname of the message server */
rfc_char_t *       group_name,  /* @parm name of a specific group of application servers */
rfc_char_t *       client,      /* @parm signon data: client */
rfc_char_t *       user,        /* @parm signon data: user id */
rfc_char_t *       password,    /* @parm signon data: password */
rfc_char_t *       language,    /* @parm signon data: language */
int          trace,       /* @parm trace ( 0 / 1 ). */
RFC_HOSTNAME as_hostname, /* @parm name of the connected App. Server (Output parm.) */
int  *       sysnr);      /* @parm system number of the connected R/3 system (Output parm.) */

/* @func
 *
 *
 * =====================================================
 *
 *   Please use <f RfcOpenEx> instead of this function
 *
 * =====================================================
 *
 *
 * opens a RFC connection.

 * all parameters are passed as single fields.
 *
 *
 * @rdesc
 * returns a handle to the RFC connection (<t RFC_HANDLE>) or
 * RFC_HANDLE_NULL.
 *
 * @xref <f RfcOpenEx>
 */
RFC_HANDLE SAP_API RfcOpenExt (
rfc_char_t*    destination,  /* @parm name of destination. */
RFC_MODE mode,         /* @parm connection mode ( <t RFC_MODE> ). */
rfc_char_t*    hostname,     /* @parm hostname of target system. */
int      sysnr,        /* @parm system number ( 0 - 99 ). */
rfc_char_t*    gateway_host, /* @parm gateway hostname or NULL */
rfc_char_t*    gateway_service,  /* @parm gateway service or NULL */
rfc_char_t*    client,           /* @parm signon data: client */
rfc_char_t*    user,             /* @parm signon data: user id */
rfc_char_t*    password,         /* @parm signon data: password */
rfc_char_t*    language,         /* @parm signon data: language */
int      trace);           /* @parm trace ( 0 / 1 ). */

/* @func
 *
 *
 * =====================================================
 *
 *   Please use <f RfcOpenEx> instead of this function
 *
 * =====================================================
 *
 *
 * opens a RFC connection.
 *
 * all parameters are passed as single fields.
 * (using the all possible <t RFC_MODE>)
 *
 * Following parameters in this call are always needed:
 *   <t RFC_MODE> mode,
 *   SAP-logon-Info: client, user, password, language
 *   trace
 *   use_sapgui
 *
 * Depending of <t RFC_MODE> mode following parameters are necessary:
 *
 *
 * RFC_MODE_PARAMETER:
 *
 *   - destination must be defined.
 *
 *   - the 'saprfc.ini' file is needed.
 *
 *   - the 'sideinfo' file is NOT necessary.
 *
 * RFC_MODE_CPIC_EXT:
 *
 *   - destination is not necessary.
 *
 *   - gateway_host, gateway_service, tp_host and tp_name must be defined.
 *
 *   - the 'saprfc.ini' and the 'sideinfo' file are NOT necessary.
 *
 * RFC_MODE_CPIC:
 *
 *   - destination must be defined.
 *
 *   - gateway_host and gateway_service can be defined.
 *
 *   - sideinfo is necessary.
 *
 * RFC_MODE_R3ONLY/RFC_MODE_VERSION_3:
 *
 *   - use_load_balancing is 0 (OFF)
 *
 *       destination is not NULL:
 *
 *         gateway_host, gateway_service, hostname and sysnr
 *         can be defined. sideinfo is necessary.
 *
 *       destination is NULL:
 *
 *         gateway_host and gateway_service can be defined.
 *         hostname, sysnr must be defined.
 *
 *   - use_load_balancing is 1 (ON):
 *
 *       destination, gateway_host, gateway_service,
 *       hostname and sysnr will not be evaluated.
 *
 *       lb_host and lb_system_name must be defined.
 *       if lb_group is NULL the group 'PUBLIC' will be used.
 *
 *
 * ATTENTION:
 *
 *   - use_sapgui and ABAP-Debug are only available with R/3 3.0C or later
 *     and not on Windows with 16-bit-RFC-library.
 *     Moreover, the functionality "INVISIBLE SAPGUI" after each RFC-function
 *     requires R/3, Rel. >= 3.0F and only available with SAPGUI,
 *     Rel. >= 3.0E on Windows NT or 95.
 *
 *   - use_sapgui and ABAP-Debug are not available with R/2.
 *
 *
 * @rdesc
 * returns a handle to the RFC connection (<t RFC_HANDLE>) or
 * RFC_HANDLE_NULL.
 *
 * @xref <f RfcOpenEx>
 */
RFC_HANDLE SAP_API RfcOpenExtV3 (
rfc_char_t*    destination,  /* @parm name of destination or NULL */
RFC_MODE mode,         /* @parm connection mode ( <t RFC_MODE> ) */
rfc_char_t*    hostname,     /* @parm hostname of target system or NULL */
int      sysnr,        /* @parm system number ( 0 - 99 ) */
rfc_char_t*    gateway_host, /* @parm gateway hostname or NULL */
rfc_char_t*    gateway_service,  /* @parm gateway service or NULL */
rfc_char_t*    client,           /* @parm signon data: client */
rfc_char_t*    user,             /* @parm signon data: user id */
rfc_char_t*    password,         /* @parm signon data: password */
rfc_char_t*    language,         /* @parm signon data: language */
int      trace,            /* @parm trace (OFF/ON/ABAP-DEBUG: 0/1/2) */
RFC_INT  use_load_balancing, /* @parm using load balancing feature (OFF/ON: 0/1)*/
rfc_char_t*    lb_host,          /* @parm Host name of the 'message server' */
rfc_char_t*    lb_system_name,   /* @parm Name of the target system (C11 i.e.) */
rfc_char_t*    lb_group,         /* @parm application server group or NULL for PUBLIC */
RFC_INT  use_sapgui        /* @parm using sapgui to display SAP dynpros and graphics
			    * (OFF/ON/INVISIBLE_SAPGUI after each RFC-function: 0/1/2)
			    */
);



#ifdef SAPwithUNICODE
#define RFC_ROLE_CLIENT        cU('C')
#define RFC_ROLE_SERVER        cU('S')

#define RFC_SERVER_R2          cU('2')
#define RFC_SERVER_R3          cU('3')
#define RFC_SERVER_EXT         cU('E')
#define RFC_SERVER_EXT_REG     cU('R')

#define RFC_PASS_UNKN          cU(' ')
#define RFC_PASS_OK            cU('O')
#define RFC_PASS_MUST_CHNG     cU('M')
#define RFC_PASS_NO_CHNG       cU('N')
#else
#define RFC_ROLE_CLIENT        'C'
#define RFC_ROLE_SERVER        'S'

#define RFC_SERVER_R2          '2'
#define RFC_SERVER_R3          '3'
#define RFC_SERVER_EXT         'E'
#define RFC_SERVER_EXT_REG     'R'

#define RFC_PASS_UNKN          ' '  /* unknown (RFC server,
                                     * or SAPsystem older than 46C
                                     * or no roundtrip yet)
                                     */
#define RFC_PASS_OK            'O'  /* OK */
#define RFC_PASS_MUST_CHNG     'M'  /* expired must be changed */
#define RFC_PASS_NO_CHNG       'N'  /* cannot be cahnged: on this SAP system,
                                     * password can only be changed once a day
                                     */

#endif

/* @func
 *
 * With this function, an RFC connection will be opened and if the RFC server
 * is an R/2 or R/3 system, it will try to login onto the according
 * SAP system. There are some advantages with this feature:
 *
 * 1. The structure of parameters in this call need not be changed for
 *    new functionalities in future.
 *
 * 2. SAP logon check at OPEN-time and not at the first call of a
 *    function module as usual.
 *
 * 3. Better performance if the RFC server is an R/3 from 3.0 onwards,
 *    because data will be compressed before transfer and not after the
 *    first call as usual.
 *
 *
 * This call should be used instead of RfcOpen, RfcOpenExt, RfcOpenExtV3
 * and RfcConnect.
 *
 *
 * @head2 Structure and contents of the 'Connect_param' |
 * This parameter is a zero terminated string and must have the following
 * structure:
 *
 * @head3 ID=value ID=value ... ID=value |
 *
 * Pay attention that the ID's are not case-sensitive but the values are
 * case-sensitive and must be set in "..." if blanks are included.
 *
 * Depend on the RFC server (R/2, R/3 or External system), some of
 * these ID's are fix and some of them are optional.
 *
 * Following ID's are available:
 *
 * TYPE    RFC server type, 2/3/E: R/2 or R/3 or External System
 *         (Default: 3)
 *
 * CLIENT  SAP logon client
 *
 * USER    SAP logon user
 *
 * PASSWD  SAP logon password
 *
 * MYSAPSSO SAP Cookie Version 1. Will be used instead user and password for logon on backend
 *
 * MYSAPSSO2 SAP Cookie Version 2. Will be used instead user and password for logon on backend
 *
 * GETSSO2 Request to create a cookie version 2 using given password and user.
 *         If the value is 1, the cookie will be generated from user and password values
 *         given by USER=user and PASSWORD=password in the same Connect_param string.
 *         If the RfcOpenEx call is succesfull, the generated cookie version 2 can be restored
 *         by using <f RfcGetTicket> API.
 *
 * LANG    SAP logon language (1-byte SAP language or 2-byte ISO language)
 *
 * CODEPAGE The given  codepage is to be used when opening a connection (Default is 
 *          either 1100 or set by <f RfcSetSystemCodepage>).
 *          Could be rather usefull if the sapgui should be started with codepage
 *          differs from 1100. 
 *
 * LCHECK  Logon check at OPEN time (0/1: without/with check, Def. 1)
 *
 * NEWPASS Changes the password during the logon
 *         !!!On SAP system kernel older than 46C the password is sent
 *         in clear text on the network!!!!
 *
 * TRACE   RFC trace (0/1: without/with trace, Def. 0)
 *
 * DEST    Destination in saprfc.ini if working with saprfc.ini.
 *         If the RFC server is an R/2 system this destination must
 *         also be defined in the 'sideinfo' for the SAP gateway.
 *
 * GWHOST  Host name of the SAP gateway (if server is R/2 or External)
 *
 * GWSERV  Service of the SAP gateway   (if server is R/2 or External)
 *
 * MSHOST  Host name of the Message Server (if using Load Balancing)
 *
 * R3NAME  Name of the R/3 system (if using Load Balancing)
 *
 * GROUP   Name of the group of application servers (if using Load Balancing)
 *
 * ASHOST  Host name of a specific application server (R/3, No Load Balancing)
 *
 * SYSNR   R/3 system number (R/3, No Load Balancing)
 *
 * TPHOST  Host name of the external RFC server program
 *
 * TPNAME  Path and name of the external RFC server program
 *         or Program ID of an registered RFC server program.
 *
 * USE_SAPGUI  RFC with SAPGUI, 0/1/2: without/with/invisible SAPGUI
 *             between two RFC functions (Default: 0). If the sapgui
 *             is to be started with codepage differs from 1100, please
 *             use option CODEPAGE to define the codepage you need.
 *
 * ABAP_DEBUG  RFC with ABAP debugger, 0/1: without/with ABAP debugger
 *             (Default: 0)
 *
 * SNC_MODE         Working with SNC, 0/1: without/with SNC
 *                  (Default: 0, s. <t RFC_SNC_MODE>)
 *
 * SNC_QOP          SNC Quality of service
 *                  (Default: 8 (RFC_SNC_QOP_DEFAULT), s. <t RFC_SNC_QOP>)
 *
 * SNC_MYNAME       Own SNC name if you don't want to use the default SNC name
 *
 * SNC_PARTNERNAME  SNC name of the SNC partner (RFC server)
 *                  or SNC name of the message server (Load Balancing)
 *
 * SNC_LIB          Path and name of the SNC-library
 *
 * SAPLOGON_ID   String defined for SAPLOGON on 32-bit Windows
 *
 * GRT_DATA      SAProuter connect data for SAPGUI when using RFC with SAPGUI.
 *               /H/......  : the whole router string for SAPGUI.
 *               /P/password: If the password for the SAPGUI connection is not
 *                              the same as the one for the RFC connection.
 *
 * @head2 EXAMPLES |
 *
 * Some examples for working with these ID's.
 *
 * @head3 1. RFC server is an R/2 system |
 *
 *    Following data are fix:  TYPE=2 DEST=... CLIENT=... USER=... PASSWD=... LANG=D
 *
 *    no additional info necessary if using saprfc.ini
 *
 *    additional info if not using saprfc.ini:   GWHOST=ihsap4 GWSERV=sapgw00
 *
 *    additional info if using SAPLOGON on 32-bit Windows:  SAPLOGON_ID=...
 *
 *    In all cases, the sideinfo for the according SAP gateway must
 *    have an entry with the destination above.
 *
 *
 * @head3 2. RFC server is an R/3 system |
 *
 *    Following data are fix:  CLIENT=... USER=... PASSWD=... LANG=D
 *
 *    additional info if using saprfc.ini    :   DEST=BIN
 *
 *    additional info if not using saprfc.ini
 *
 *      + Connect to a specific application server: ASHOST=hs0311 SYSNR=53
 *
 *      + Connect using Load Balancing: MSHOST=hs0311 R3NAME=BIN GROUP=PUBLIC
 *
 *        If no info about MSHOST, the RFC library will try to get this
 *        host name in the 'sapmsg.ini' customized for SAPLOGON on Windows
 *        (in Windows-directory). This file can also be copied and used on
 *        a Non-Windows-Host. It must be in the working directory of the RFC
 *        client or its path name must be defined via the environment
 *        'RFC_LOGON_INI_PATH' (e.g. /usr/rfctest).
 *        The SAProuter string, if exists, must be defined in the R/3 name.
 *
 *        e.g.  R3NAME=/H/sapgate1/S/3297/LOI where LOI is the R/3 system name.
 *
 *        or    R3NAME=/SAP_test/LOI    or    R3NAME='/SAP America/LOI'
 *              where 'SAP_test' and 'SAP America' is the router name defined
 *              in the 'saproute.ini' for SAPLOGON on Windows.
 *
 *    additional info if using SAPLOGON on 32-bit Windows:  SAPLOGON_ID=...
 *
 *
 * @head3 3. RFC server is an external program |
 *
 *    Following data are fix:  TYPE=E
 *
 *    additional info if using saprfc.ini    :   DEST=RFCEXT
 *
 *    additional info if not using saprfc.ini
 *
 *      In this case, the info about TPHOST determines whether your RFC
 *      client will connect to an RFC server started by the SAP gateway
 *      or it will connect to an already registered RFC server.
 *
 *      + RFC server will be started by the SAP gateway
 *
 *        GWHOST=hs0311 GWSERV=sapgw53 TPHOST=hs0311 TPNAME=/sap/srfcserv
 *
 *        If GWHOST and TPHOST are different, make sure that the SAP gateway
 *        has the access to start the server program via remote shell.
 *
 *
 *      + RFC server is already registered at the SAP gateway
 *
 *        GWHOST=hs0311 GWSERV=sapgw53 TPNAME=hw1145.srfcserv
 *
 *
 * See the delivered program srfctest.c for more details.
 *
 *
 *
 * @head2 ATTENTION |
 *
 *   - USE_SAPGUI and ABAP_DEBUG are only available with R/3 3.0C or later
 *     and not on Windows with 16-bit-RFC-library.
 *     Moreover, the functionality "INVISIBLE SAPGUI" after each RFC-function
 *     requires R/3 from 3.0F onwards and only available with SAPGUI from 3.0E
 *     onwards on Windows NT/95 or with SAPGUI from 3.0F onwards on UNIX.
 *
 *   - USE_SAPGUI and ABAP_DEBUG are not available with R/2.
 *
 *   - SAPLOGON_ID is only available on 32-bit Windows where SAPLOGON is also installed.
 *     Using this feature, all parameters needed for open a connection to an SAP system
 *     such as DEST, GWHOST, GWSERV, ASHOST, SYSNR, MSHOST, R3NAME, GROUP and SNC-parameters
 *     are not necessary because the RFC library will get this information from the
 *     SAPLOGON data files.
 *
 *
 * @rdesc
 * returns a valid handle to the RFC connection (<t RFC_HANDLE>) or
 * RFC_HANDLE_NULL if connection could not be established.
 *
 * @comm
 * 1. Every connection will be established either
 * with default code page 1100 or 
 * according environment variable SAP_CODEPAGE or
 * according id CODEPAGE in <p connect_param> string of this function.
 *
 * 2. Do not call <f RfcLastErrorEx> if RFC_HANDLE_NULL was returned but
 * extract the error information from <p error_info>.
 *---------------------------------------------------------------------------*/
RFC_HANDLE SAP_API RfcOpenEx
(
 rfc_char_t         * connect_param, /* @parm connection parameters.*/
 RFC_ERROR_INFO_EX  * error_info     /* @parm <t RFC_ERROR_INFO_EX> describing the error.*/
);


/* @func
 *
 *
 * =====================================================
 *
 *   Please use <f RfcOpenEx> instead of this function
 *
 * =====================================================
 *
 *
 * opens a connection according to the given options.
 *
 * The structure <t RFC_OPTIONS>
 * contains the necessary data for opening the connection
 *
 *
 * @rdesc
 * returns a handle to the RFC connection (<t RFC_HANDLE>) or
 * RFC_HANDLE_NULL, if the connection cannot be opened.
 *
 * @xref <f RfcOpenEx>
 */
RFC_HANDLE SAP_API RfcOpen(
  RFC_OPTIONS * options  /* @parm connection parameters as described
                          * at structure <t RFC_OPTIONS>.
                          */
                          );



/*--------------------------------------------------------------
 * @func
 * accepts an incoming connection and has to be used, if the program
 * was started by a RFC call issued by a SAP system or an other
 * program using this API. The command line (argv) has to be passed
 * to this function.
 *
 * With Release 3.0C onwards this function can be used to register at
 * a SAP gateway and the server program can wait for next RFC request
 * by issueing <f RfcDispatch> or <f RfcWaitForRequest> or <f RfcListen>
 * or <f RfcGetName>.
 *
 * Using this functionality a RFC server program can now already run
 * and work as a RFC daemon. A start of any server program by a R/3
 * application server, by SAPGUI or via remote shell by a SAP gateway
 * is no longer necessary.
 *
 * There are 2 ways to define the input parameter 'argv':
 *
 * 1. Working with 'saprfc.ini'-file:
 *
 *    -D<lt>destination pointed to an entry in 'saprfc.ini'<gt>
 *
 *    -t      Set RFC-Trace on
 *
 *
 * 2. Working without 'saprfc.ini'-file:
 *
 *    -a<lt>program ID<gt> e.g. own_host_name.program_name
 *
 *    -g<lt>host name of the SAP gateway<gt>
 *
 *    -s<lt>service of the SAP gateway<gt>, e.g. sapgw00
 *
 *    -t      Set RFC-Trace on
 *
 *
 *
 * We recommend the first one because you can use some RFC features today
 * and in the future without changes in your RFC programs.
 *
 * See saprfc.ini for more details.
 *
 *
 *
 * The 3 parameters above must fit with the configuration in R/3.
 * (via SM59, Connection type T and Register Mode)
 *
 * For working with register mode there are no other changes in RFC
 * server programs necessary.
 *
 * Be carefull by using only the program name as program ID because
 * there might be more programs (e.g. rfcexec) registered at the
 * same SAP gateway. Please use at least the host name of the computer where
 * your RFC server program is running as a part of the program ID.
 *
 * Example: The well-known rfcexec program can be started from the
 *          command line with RFC-trace as follows:
 *
 *          rfcexec -ap10234.rfcexec -ghs0311 -xsapgw53 -t
 *
 *       or
 *
 *          rfcexec -Drfctest
 *
 *          and an entry in saprfc.ini can be defined as follows:
 *
 *          DEST=rfctest
 *
 *          TYPE=R
 *
 *          PROGID=p10234.rfcexec
 *
 *          GWHOST=hs0311
 *
 *          GWSERV=sapgw53
 *
 *          RFC_TRACE = 1
 *
 *
 *
 * RFC server programs working with this functionality will be called
 * RFC server programs running in register mode.
 *
 *
 * If an RFC server program is working with SNC (Secure Network
 * Communications) it can be started with the following SNC parameters:
 *
 *    -L<lt>SNC library name on the local system<gt> e.g. /krb5/hpux/lib/libkrb5.sl
 *
 *    -S<lt>SNC name of this local instance<gt>
 *
 *
 * ATTENTION:
 *
 * RFC program can use <f RfcCheckRegisterServer> to test whether
 * and how many server programs are registered at a SAP gateway with a
 * defined program ID. Moreover, it can use <f RfcCancelRegisterServer>
 * to cancel all RFC servers which are registered at a SAP gateway with a
 * defined program ID.
 *
 *
 * @rdesc
 * returns a valid <t RFC_HANDLE> or RFC_HANDLE_NULL
 *--------------------------------------------------------------
 */
RFC_HANDLE SAP_API RfcAccept(
  rfc_char_t ** argv /* @parm command line (argv),
                * the program is started with.
                */
  );

RFC_HANDLE SAP_API RfcAcceptExt( rfc_char_t* CommandLine );

/*----------------------------------------------------------------------------
@func

closes a RFC connection or all RFC connections.

----------------------------------------------------------------------------*/
void       SAP_API RfcClose(
  RFC_HANDLE handle   /* @parm handle to RFC connection. If handle is RFC_HANDLE_NULL, all
                         opened connections will be closed.
                       */
                           );

/*----------------------------------------------------------------------------
@func

sends a error message, if possible, and closes the connection.

A given error message cannot be send, if the receiver is not in state,
where it expects some RFC data to receive.
----------------------------------------------------------------------------*/
void       SAP_API RfcAbort(
  RFC_HANDLE handle,  /* @parm handle to RFC connection */
  rfc_char_t * text   /* @parm error message.
                       *
                       * If this field is NULL, the connection is
                       * only closed.
                       */
  );

/* @enum RFC_RC | RFC return codes */
typedef enum
{
   RFC_OK,                /* @emem O.K.
                           */
   RFC_FAILURE,           /* @emem Error occurred
                           */
   RFC_EXCEPTION,         /* @emem Exception raised
                           */
   RFC_SYS_EXCEPTION,     /* @emem System exception raised, connection closed
                           */
   RFC_CALL,              /* @emem Call received
                           */
   RFC_INTERNAL_COM,      /* @emem Internal communication, repeat (internal use only)
                           */
   RFC_CLOSED,            /* @emem Connection closed by the other side
                           */
   RFC_RETRY,             /* @emem No data yet (RfcListen or RfcWaitForRequest only)
                           */
   RFC_NO_TID,            /* @emem No Transaction ID available
                           */
   RFC_EXECUTED,          /* @emem Function already executed
                           */
   RFC_SYNCHRONIZE,       /* @emem Synchronous Call in Progress (only for Windows)
                           */
   RFC_MEMORY_INSUFFICIENT,/* @emem Memory insufficient
                           */
   RFC_VERSION_MISMATCH,  /* @emem Version mismatch
                           */
   RFC_NOT_FOUND,         /* @emem Function not found (internal use only)
                           */
   RFC_CALL_NOT_SUPPORTED,/* @emem This call is not supported
                           */
   RFC_NOT_OWNER,         /* @emem Caller does not own the specified handle
                           */
   RFC_NOT_INITIALIZED,   /* @emem RFC not yet initialized
                           */
   RFC_SYSTEM_CALLED,     /* @emem A system call such as RFC_PING for connection
			   * test is executed
                           */
   RFC_INVALID_HANDLE,    /* @emem An invalid handle was passed to an API call
                           */
   RFC_INVALID_PARAMETER, /* @emem An invalid parameter was passed to an API call
			   */
   RFC_CANCELED           /* @emem Internal use only
                          */
}
RFC_RC;

/* @struct RFC_PARAMETER |
 * RFC parameters.
 *
 * Describes 'Exporting' and 'Importing' parameters of the
 * interface of a function module.
 *
 * @xref <t RFC_CHAR>, <t RFC_NUM>, <t RFC_BYTE>, <t RFC_BCD>,
 *       <t RFC_INT1>, <t RFC_INT2>, <t RFC_INT>, <t RFC_FLOAT>,
 *       <t RFC_DATE>, <t RFC_TIME>, <t RFC_STRING>.
 */
typedef struct
{
   void *   name;  /* @field Name of the field (in the interface
                    * definition of the function).
                    */
   unsigned nlen;  /* @field Length of the name
                    * (should be strlen(name)).
                    */
   RFC_TYPEHANDLE type;  /* @field Data type of the field.
                          * In former versions this field was simply
                          * defined as 'unsigned'.
                          */
   unsigned leng;  /* @field Length of the field in Bytes.
                    */
   void *   addr;  /* @field Address of the field to be exported
                    * or imported.
                    */
}
RFC_PARAMETER;


#ifndef SAPITAB_H
typedef void * ITAB_H;
#endif

/* @enum RFC_ITMODE | mode how to pass internal table */
typedef enum
{
   RFC_ITMODE_BYREFERENCE,  /* @emem Table is passed by reference
                             * (ALLWAYS USE THIS)
                             */
   RFC_ITMODE_BYVALUE,      /* @emem Table is passed by value, changes
                             * are not transported back
                             * (internal use only !)
                             */
   RFC_ITMODE_KEEPALIVE     /* @emem Table is passed by reference, but
                             * is kept alive after returning,
                             * (i.e. after RfcSendData, internal use
                             * only )
                             */
}
RFC_ITMODE;

/* @struct RFC_TABLE |
 * RFC tables.
 *
 * Describes Tables parameters of the interface of a function module.
 */
typedef struct
{
   void *   name;      /* @field Name of the table (in the interface
                        * definition of the function).
                        */
   unsigned nlen;      /* @field Length of the name
                        * (should be strlen(name)).
                        */
   RFC_TYPEHANDLE type;      /* @field Data type of the lines of the table.
                              * In former versions this field was simply
                              * defined as 'unsigned'.
                              */
   unsigned leng;      /* @field length of a row in bytes.
                        */
   ITAB_H   ithandle;  /* @field Table handle <t ITAB_H>,
                        * i.e. the address of
                        * the control structure of the
                        * internal table.
                        *
                        * This is an input field at <f RfcCall> and
                        * an output field at <f RfcGetData>.
                        */
   RFC_ITMODE itmode;  /* @field mode ( <t RFC_ITMODE> ),
                        * how this table has to be received :
                        * 'call by reference' or 'call by value'.
                        * (only for RfcGetData, has no effect
                        *  in RfcCall).
                        */
   int      newitab;   /* @field If the value is not 0 after RfcGetData,
                        * the table was created by RfcGetData.
                        * Internal use only. This field must not
                        * be modified
                        * between <f RfcGetData> and <f RfcSendData> !
                        */
}
RFC_TABLE;




/*------------------------------------------------------
 *
 * @func
 *
 *  Function to allocate, reallocate or free a <t RFC_BLOB>.
 *
 *  @rdesc
 *  returns <e RFC_RC.RFC_OK> if no error occurred and
 *  <e RFC_RC.RFC_FAILURE> elsewhere.
 *
 *  @xref <f RFC_BLOB>
 *
 *------------------------------------------------------*/
RFC_RC SAP_API RfcResizeBlob
(
   RFC_BLOB * pBlob, /* @parm pointer to BLOB. */
   unsigned   leng   /* @parm new size of BLOB. If leng is 0,
		      * the BLOB is freed.
		      */
);



/*----------------------------------------------------------------------------
 * @func
 *
 * The RFC library determines the own codepage for every RFC
 * connection at open time. 
 *
 * Every connection will be established either
 * with default code page (1100) or 
 * according environment variable SAP_CODEPAGE or
 * according option CODEPAGE in <f RfcOpenEx> or according
 * new default codepage set by <f RfcSetSystemCodePage>.
 * 
 * With this call, a RFC client or server program can set an own
 * codepage for a RFC connections to runtime. In this case, this call must 
 * be issued after <f RfcOpen>, <f RfcOpenEx> or <f RfcAccept>. 
 * 
 * Alternatively one can use option CODEPAGE
 * of <f RfcOpenEx> to define the codepage for a RFC connection to open time. 
 * 
 * If one wants to change the default codepage, <f RfcSetSystemCodePage> must
 * be called.
 * 
 * @xref <f RfcOpenEx>, <f RfcOpen>, <f RfcSetSystemCodepage>, <f RfcCharConversion>.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull
 *
 * <e RFC_RC.RFC_INVALID_HANDLE> if handle is invalid
 *
 * <e RFC_RC.RFC_NOT_OWNER> if caller does not own the specified handle
 *
 * <e RFC_RC.RFC_NOT_INITIALISED> if initialisation has not be done or failed
 *----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcSetCodePage
(
  RFC_HANDLE   handle,      /* @parm handle to RFC connection. */
  rfc_char_t * codepage     /* @parm own code page (e.g. "1800") */
);




/*----------------------------------------------------------------------------
 * @func
 * 
 * With this call, a RFC client or server process can change the default codepage.
 * The call may be called to any time. All connections opened after this call,
 * will be established with new default codepage.
 * 
 * Alternatively one can use option CODEPAGE
 * of <f RfcOpenEx> to set the codepage for a RFC connection 
 * at open time (in this case the default codepage will be ignored).
 * 
 * @xref <f RfcOpenEx>, <f RfcOpen>, <f RfcSetCodepage>, <f RfcCharConversion>.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull
 *
 * <e RFC_RC.RFC_NOT_INITIALISED> if initialisation was not be done or failed.
 *
 * @comm 
 * - Please notice that the codepage's change is valid only for all connections,
 * which will be established after this call. All already existing connections
 * are not concerned.
 *
 * - Every connection, which will be established after this call, will be done either
 * with new default code page or according option CODEPAGE in <f RfcOpenEx>.
 *----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcSetSystemCodePage 
(
  rfc_char_t * codepage  /* @parm codepage (e.g "8000") */
);


/*----------------------------------------------------------------------------
@func

An RFC program can use this call to check whether and how
many RFC server programs are registered at a SAP gateway with a defined
program ID.

This call is only available with a SAP gateway from 3.1G onwards.

An RFC server program which registers at a SAP gateway and then waits for
RFC requests from any R/2 or R/3 or another external program can have
one of the following 3 states:

INIT:  RFC server is only registered at the SAP gateway, it doesn't wait
       at this moment for RFC request. An RFC server enters in this state
       after <f RfcAccept> in registered mode.

READY: RFC server enters in this state after <f RfcListen>,
       <f RfcWaitForRequest>, <f RfcGetName> or <f RfcDispatch>.
       It is now ready to process incoming RFC requests. Be careful,
       RfcGetName and RfcDispatch are blocking calls. They will
       be returned when an RFC request is incoming.
       Make sure that the time between two RfcListen or RfcWaitForRequest
       is not longer than the CPIC_TIMEOUT defined for the SAP gateway.
       This is necessary because the SAP gateway will only queue the RFC
       request withhin the CPIC time above. If this time is elapsed the
       RFC request will be returned back to the RFC client with an
       according error.

BUSY:  RFC server is processing an RFC request and is not available
       for other ones.

@xref <f RfcCancelRegisterServer>.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcCheckRegisterServer(
  rfc_char_t      * tpid,       /* @parm program ID of the registered RFC server program */

  rfc_char_t      * gwhost,     /* @parm host name of the SAP gateway */
  
  rfc_char_t      * gwserv,     /* @parm service of the SAP gateway */
  
  RFC_INT         * ntotal,     /* @parm total number of all registered RFC server programs
                                 *  (Sum of ninit, nready and nbusy)
                                 */
  RFC_INT         * ninit,      /* @parm number of registered RFC server programs
                                 * in INIT-state with the program ID above.
                                 */
  RFC_INT         * nready,     /* @parm number of registered RFC server programs
			                     * in READY-state with the program ID above.
			                     */
  RFC_INT         * nbusy,      /* @parm number of registered RFC server programs
                                 * in BUSY-state with the program ID above.
                                 */
  RFC_ERROR_INFO_EX * error_info    /* @parm structure
				                     * <t RFC_ERROR_INFO_EX> describing the error
                                     */
  );

/*----------------------------------------------------------------------------
@func

An RFC program can use this call to cancel all registered RFC servers
at a SAP gateway with a defined program ID.

Pay attention that RFC servers in BUSY state cannot be canceled.
In this case you will have a difference between the output parameters
ntotal and ncancel returned by this call.

This call is only available with a SAP gateway from 3.1H onwards.


@xref <f RfcCheckRegisterServer>.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcCancelRegisterServer(
  rfc_char_t      * tpid,       /* @parm program ID of the registered RFC server program
			   */
  rfc_char_t      * gwhost,     /* @parm host name of the SAP gateway
			   */
  rfc_char_t      * gwserv,     /* @parm service of the SAP gateway
			   */
  RFC_INT   * ntotal,     /* @parm total number of all registered RFC server programs
			   */
  RFC_INT   * ncancel,    /* @parm total number of all canceled RFC server programs
			   */
  RFC_ERROR_INFO_EX * error_info    /* @parm structure
				     * <t RFC_ERROR_INFO_EX> describing the error
                                     */
  );


/*----------------------------------------------------------------------------
 * @func RfcCleanupContext
 *
 * Cleanup the attached R/3 context
 *
 * Resets the state aquired by the R/3 user context as well as
 * local cache information.  This call is intended for client programs
 *
 * This is only usefull when running against an R/3 reslease 40A
 * and above.
 *
 * If the remote system is not an R/3 release 40A and above,
 * RfcCleanupContext calls <f RfcClose> internally and returns
 * RFC_CLOSE,
 *
 * @rdesc
 * <e RFC_RC.RFC_OK>      The context has been reset.
 * <e RFC_RC.RFC_CLOSED>  The connection has been closed internally
 *                        The caller shall call RfcOpen* to obtain a
 *                        new handle.
 * <e RFC_RC.RFC_FAILURE> An error occured. The connection is lost.
 *----------------------------------------------------------------*/
RFC_RC SAP_API RfcCleanupContext
(
 RFC_HANDLE handle  /* @parm handle to connection that must be cleaned */
);




/*---------------------------------------------------------------
 * @enum RFC_PASS_RC | ON_RFC_PASSWORD_EXPIRED reason
 * This value is passed as reaon code to the ON_RFC_PASSWORD_EXPIRED
 * call back.  This is helpfull to handle system policies
 *---------------------------------------------------------------*/
typedef enum
{
   RFC_PASS_EXPIRED,      /* @emem First call: password is expired
                           */
   RFC_PASS_REFUSED       /* @emem nth call: Password supplied
                           *       by the previous call refused
                           *       (too short, had allready been
                           *       used prevously, not enough
                           *       upercase, not enought digits...)
                           *
                           */
} RFC_PASS_RC;


/*-----------------------------------------------------------------------
 * @cb  RFC_RC |  ON_RFC_PASSWORD_EXPIRED |
 *
 * function to be installed by <f RfcRegisterPasswordChanger>.
 *
 * The function is called when an expired password is reported
 * by R/3 and when LCHECK=1 was specified in <f RfcOpenEx>
 * (this is the default).
 *
 * The function shall write the new password in the buffer
 * pointed by <p NewPassword>.
 * The buffer is guarantied to be <p cntBytes> long.
 *
 * The function may return an empty string (filled with zeros) to signal that
 * the user does not wish to change the password.
 *
 * @comm
 * Only SAP system running a kernel 46C and higher will report expired password.
 */
typedef RFC_RC (DLL_CALL_BACK_FUNCTION_PTR ON_RFC_PASSWORD_EXPIRED)
(
 RFC_PASS_RC    Reason,         /* @parm [in]  Reason for this call    */
 RFC_CHAR     * NewPassword,    /* @parm [out] The new password to use */
 RFC_INT        cntBytes        /* @parm [in]  The size of NewPassword */
);

/*----------------------------------------------------------------------------
@func

Registers a function that will be called during <f RfcOpenEx> if the
systems reports that the password is expired.

If RfcRegisterPasswordChanger() is not called, then <f RfcOpenEx> will
not report expired password.

The function is called when an expired password is reported
by R/3 and when LCHECK=1 was specified in <f RfcOpenEx>
(this is the default).


Only SAP system running a kernel 46C and higher will report expired password.

@xref <f RfcCancelRegisterServer>.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.
*/
RFC_RC SAP_API RfcRegisterPasswordChanger
(
 ON_RFC_PASSWORD_EXPIRED pCbChangePassword /* @parm a call back function that will be called
                                            * if a SAP system reports an expired password
                                            */
        );




/*----------------------------------------------------------------------------
@func

installs a description for a structure being used in an RFC interface.

For functions with structured parameters or tables it is necessary to
install a description of the structures used to allow automatic conversion
of different data representations by the RFC library.

The description must contain all (scalar) fields of the structure in correct
order. A type handle is returned which can be used in <t RFC_PARAMETER> and
<t RFC_TABLE> descriptions.

@rdesc
  returns
  @flag <e RFC_RC.RFC_OK>  |
        The structure was successfully installed. The returned
        type handle (<t RFC_TYPEHANDLE>) can be used in
        <t RFC_PARAMETER> or <t RFC_TABLE> arrays.
  @flag <e RFC_RC.RFC_MEMORY_INSUFFICIENT> |
        not enough memory.

----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcInstallStructure(
    rfc_char_t        * name,  /* @parm the name of the structure. Is used only
                          * in trace file.
                          */
    RFC_TYPE_ELEMENT * elements, /* @parm description of the elements of
                                  * the structure
                                  */
    unsigned      entries,		 /* @parm count of (scalar) elements in
								  * the structure. 'elements' points to
                                  * an array of size
                                  * 'entries' * sizeof(RFC_TYPE_ELEMENT).
                                  */
    RFC_TYPEHANDLE * pTypeHandle /* @parm returned type handle. */
);


/*----------------------------------------------------------------------------
 * @func
 *
 * same as <f RfcINstallStructure> but uses the information returned by the
 * function module RFC_GET_STRUCTURE_DEFINITION
 *
 * This is helpful for structures using .INCLUDE in the data dictionary.
 *
 *----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcInstallStructure2
(
    rfc_char_t        * name,   /* @parm the name of the structure. Is used only
								 * in trace file.
								 */
    RFC_TYPE_ELEMENT2 * elements, /* @parm description of the elements of
                                  * the structure
                                  */
    unsigned      entries,  /* @parm count of (scalar) elements in
                             * the structure. 'elements' points to
                             * an array of size
                             * 'entries' * sizeof(RFC_TYPE_ELEMENT).
                             */
    RFC_TYPEHANDLE * pTypeHandle /* @parm returned type handle. */
                          );


/* -------------------------------------------------------------------
 * security and automatic (re)connection
 * ------------------------------------------------------------------*/

typedef RFC_BYTE RFC_TICKET_CHECKSUM[20];   /* SHA checksum */
typedef RFC_BYTE RFC_TICKET_SYSNAME[8];

typedef struct
{
                                        /* position     */
   RFC_BYTE     head    [4];            /* 0-4          */
   RFC_BYTE     unused  [8];            /* 4-12         */
   RFC_BYTE     ip      [4];            /* 12-16        */
   RFC_BYTE     sysnr   [1];            /* 16-17        */
   RFC_BYTE     client  [3];            /* 17-20        */
   RFC_TICKET_SYSNAME   sysname;        /* 20-28        */
   RFC_BYTE     userid  [14];           /* 28-42        */
   RFC_BYTE     valid   [2];            /* 42-44        */
   RFC_TICKET_CHECKSUM  check;          /* 44-64        */
                                        /* size 64      */
}
RFC_TICKET;

/* for internal use only */
#define NEWITAB_SHARE   0
#define NEWITAB_POP     1
#define NEWITAB_KEEP    2
#define NEWITAB_BYVALUE 3
#define NEWITAB_COPY    4


/* @type RFC_FUNCTIONNAME | RFC function name.
 * Character field big enough containing the name of any
 * ABAP/4 function module.
 */
#define RFC_FUNCTIONNAME_LN   31
typedef rfc_char_t RFC_FUNCTIONNAME[RFC_FUNCTIONNAME_LN];


/* RFC Transaction Identifier (TID) */
#define RFC_TID_LN            24        /* Offsetof (ARFCSDATA, arfcdest)    */
#define RFC_CALLID_LN         64        /* Offsetof (ARFCSDATA, arfcblcnt)   */
#define ARFCSDATA_HEADER_LN   68        /* Offsetof (ARFCSDATA, arfcdata01)  */

/* @type RFC_TID |
 * zero terminated string giving  an RFC transaction a globally unique
 * identifier.
 */
typedef rfc_char_t RFC_TID[RFC_TID_LN+1];

#define RFC_QNAME_LN          24
#define RFC_QCOUNT_LN         24

/* @type RFC_QNAME |
 * zero terminated string giving a queue name for using the receive queue in
 * SAP system.
 */
typedef rfc_char_t RFC_QNAME[RFC_QNAME_LN+1];


/* ------------------------------------------------------------------
 * RFC function interface for working with SNC
 * ----------------------------------------------------------------*/

/*----------------------------------------------------------------------
 * @func
 * After being started, an RFC server program can issue this call
 * to check whether it is running in an SNC environment.
 *
 * If working with SNC an RFC server program can issue
 * <f RfcSncPartnerName> or <f RfcSncPartnerAclKey> to get the
 * SNC name or the SNC ACL key of the RFC client and can then
 * decide to accept the communication with this partner.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull.
 * <e RFC_RC.RFC_FAILURE> if any error has occured.
 *
 */
RFC_RC SAP_API RfcSncMode(
 RFC_HANDLE       handle,          /* @parm connection handle.
                                    */
 RFC_SNC_MODE *   snc_mode         /* @parm SNC mode for working with or without SNC.
				    *
                                    */
                      );


/*----------------------------------------------------------------------
 * @func
 * If the <t RFC_SNC_MODE> is on, an RFC server program can issue
 * this call to get the SNC name of the RFC client.
 * It can then decide itself whether it can communicate with this
 * RFC client.
 *
 * An RFC server program can only
 * issue this call after a successful return of <f RfcListen> or
 * <f RfcWaitForRequest> or <f RfcGetName> or <f RfcDispatch>.
 *
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if successfull.
 *
 * <e RFC_RC.RFC_FAILURE> if any error has occured.
 *
 *----------------------------------------------------------------*/
RFC_RC SAP_API RfcSncPartnerName
(
 RFC_HANDLE       handle,                 /* @parm connection handle.*/
 rfc_char_t *     partner_name,           /* @parm buffer for SNC name of
                                           * the RFC client.
                                           */
 RFC_INT          maxlen                  /* @parm Max. length of buffer.*/
);




/*----------------------------------------------------------------------
 * @func
 * Instead of getting the SNC name an RFC server program can issue
 * this call to get the SNC ACL key of the RFC client.
 *
 * See <f RfcSncPartnerName> for more details.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull
 * <e RFC_RC.RFC_FAILURE> else.
 *
 */
RFC_RC SAP_API RfcSncPartnerAclKey(
 RFC_HANDLE       handle,          /* @parm connection handle.
                                    */
 RFC_BYTE *       partner_aclkey,  /* @parm buffer for SNC ACL key of
				    * the RFC client.
                                    */
 RFC_INT          maxlen,          /* @parm Max. length of buffer.
                                    */
 RFC_INT *        aclkey_len       /* @parm Real length of the SNC ACL key.
                                    */
                      );

/*----------------------------------------------------------------------
 * @func
 * An administrativ program can issue this call to convert an SNC name
 * in an SNC ACL key without working with RFC.
 *
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull
 * <e RFC_RC.RFC_FAILURE> else.
 *
 */
RFC_RC SAP_API RfcSncNameToAclKey(
 rfc_char_t            * snc_lib,         /* @parm name of the SNC library
				     * on the local system.
                                     */
 rfc_char_t            * snc_name,        /* @parm SNC name to be converted
                                     */
 RFC_BYTE        * snc_aclkey,      /* @parm buffer for SNC ACL key
                                     */
 RFC_INT           maxlen,          /* @parm Max. length of buffer.
                                     */
 RFC_INT         * aclkey_len,      /* @parm Real length of the SNC ACL key.
                                     */
 RFC_ERROR_INFO_EX * error_info     /* @parm structure
				     * <t RFC_ERROR_INFO_EX> describing the error
                                     */
                      );

/*----------------------------------------------------------------------
 * @func
 * An administrativ program can issue this call to convert an
 * SNC ACL key in an SNC name without working with RFC.
 *
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call successfull
 * <e RFC_RC.RFC_FAILURE> else.
 *
 */
RFC_RC SAP_API RfcSncAclKeyToName(
 rfc_char_t            * snc_lib,         /* @parm name of the SNC library
				     * on the local system.
                                     */
 RFC_BYTE        * snc_aclkey,      /* @parm SNC ACL key to be converted
                                     */
 RFC_INT           aclkey_len,      /* @parm Length of the SNC ACL key.
                                     */
 rfc_char_t            * snc_name,        /* @parm SNC name to be converted
                                     */
 RFC_INT           maxlen,          /* @parm Max. length of buffer.
                                     */
 RFC_ERROR_INFO_EX * error_info     /* @parm structure
				     * <t RFC_ERROR_INFO_EX> describing the error
                                     */
                      );



/* ------------------------------------------------------------------
 * Standard RFC interface
 * ----------------------------------------------------------------*/


/* -----------------------------------------
 * calling function modules
 * ---------------------------------------*/

/* @func
 * Calls an ABAP/4 function module via RFC.
 *
 * The structures <t RFC_PARAMETER> and  <t RFC_TABLE> contain
 * name and description of the 'exporting' parameters and tables
 * (internal ABAP/4 table) of the function's interface.
 * The function returns after the call request is send.
 * If the function returns RFC_OK, there is no guaranty, that
 * the call was already received by the target system.
 *
 * @rdesc
 * returns RFC_OK or RFC_FAILURE.
 *
 * @ex calling the function module  RFC_SYSTEM_INFO. |
 *
 *   RFC_RC             rfc_rc;
 *   RFC_PARAMETER      exporting[32];
 *   RFC_TABLE          tables[32];
 *
 *   exporting[0].name = NULL;
 *   tables[0].name    = NULL;
 *
 *   rfc_rc = RfcCall( handle, "RFC_SYSTEM_INFO",
 *                                  exporting,
 *                                  tables );
 *   if( rfc_rc != RFC_OK )
 *   {
 *      ....
 *
 */
RFC_RC SAP_API RfcCall(
 RFC_HANDLE       handle,    /* @parm connection handle.
                              */
 rfc_char_t *     function,  /* @parm function module to call.
                              */
 RFC_PARAMETER *  parameters,/* @parm 'exporting' parameters.
                              */
 RFC_TABLE     *  tables     /* @parm 'tables' parameters.
                              */
                      );

/*----------------------------------------------------------------------
 * @func
 * receives the return values from a function call issued by <f RfcCall>.
 *
 * allows to receive the answer to a RFC call and must be called
 * after RfcCall was issued. The tables' description ( <t RFC_TABLE> )
 * must be identical to the one used in RfcCall.
 *
 * The function RfcReceive waits till the answer is received.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK>  The call was successfully completed and
 *       the values of the returned parameters were filled into the
 *       fields being supplied by the <t RFC_PARAMETER> array.
 * <e RFC_RC.RFC_FAILURE> An internal error has occurred.
 *       <f RfcLastError> or <f RfcLastErrorEx> may give more information.
 * <e RFC_RC.RFC_EXCEPTION>  The callee has raised an exception.
 *       The field '*exception' points to the name of the exception. No
 *       data were transported.
 * <e RFC_RC.RFC_SYS_EXCEPTION>  The local or remote RFC system has
 *       raised an exception. Also '*exception' points to the name of
 *       the exception. The connection was automatically closed by the
 *       system and <f RfcLastError> or <f RfcLastErrorEx> may give more
 *       information on the origin
 *       of the error. Two exceptions may occur now: SYSTEM_FAILURE and
 *       COMMUNICATION_FAILURE.
 * <e RFC_RC.RFC_CALL>  The callee has issued a RFC call to the
 *       caller of RfcReceive. No data are transported. The call request
 *       must be handled by using the functions <f RfcDispatch> or by
 *       <f RfcGetName>, <f RfcGetData> and <f RfcSendData>
 *       before another call to RfcReceive can be done.
 *
 * @xref <f RfcCall> <f RfcCallReceive>.
 */
  RFC_RC SAP_API RfcReceive(
  RFC_HANDLE    handle,      /* @parm connection handle.
                              */
  RFC_PARAMETER *  parameters,/* @parm 'importing' parameters.
                               */
  RFC_TABLE     *  tables,   /* @parm 'tables' parameters.
                              */
  rfc_char_t **          exception /* @parm output parameter:
                              * pointer to exception string.
                              * only set, if <e RFC_RC.RFC_EXCEPTION> or
                              * <e RFC_RC.RFC_SYS_EXCEPTION> is returned.
                              */
 );


/* @func
 * calls an ABAP/4 function module and receives the return values
 * in one function call.
 * @rdesc
 * returns
 * @flag <e RFC_RC.RFC_OK>  |
 *       The call was successfully completed and
 *       the values of the returned parameters were filled into the
 *       fields being supplied by the <t RFC_PARAMETER> array.
 * @flag <e RFC_RC.RFC_FAILURE> |
 *       An internal error has occurred.
 *       <f RfcLastError> or <f RfcLastErrorEx> may give more information.
 * @flag <e RFC_RC.RFC_EXCEPTION> |
 *       The callee has raised an exception.
 *       The field '*exception' points to the name of the exception. No
 *       data were transported.
 * @flag <e RFC_RC.RFC_SYS_EXCEPTION> |
 *       The local or remote RFC system has
 *       raised an exception. Also '*exception' points to the name of
 *       the exception. The connection was automatically closed by the
 *       system and <f RfcLastError>  or <f RfcLastErrorEx> may give more
 *       information on the origin
 *       of the error. Two exceptions may occur now: SYSTEM_FAILURE and
 *       COMMUNICATION_FAILURE.
 * @flag <e RFC_RC.RFC_CALL>  |
 *       The callee has issued an RFC call to the
 *       caller of RfcReceive. No data are transported. The call request
 *       must be handled by using the functions <f RfcDispatch> or by
 *       <f RfcGetName>, <f RfcGetData> and <f RfcSendData>
 *       before an other call to RfcReceive can be done.
 *
 * @xref <f RfcCall> <f RfcReceive>.
 */
RFC_RC SAP_API RfcCallReceive(
 RFC_HANDLE       handle,    /* @parm connection handle.
                              */
 rfc_char_t    *  function,  /* @parm function module to call.
                              */
 RFC_PARAMETER *  exporting, /* @parm 'exporting' parameters.
                              */
 RFC_PARAMETER *  importing,/* @parm 'importing' parameters.
                              */
 RFC_TABLE     *  tables,   /* @parm 'tables' parameters.
                             */
 rfc_char_t   **  exception /* @parm output parameter:
                             * pointer to exception string.
                             * only set, if <e RFC_RC.RFC_EXCEPTION> or
                             * <e RFC_RC.RFC_SYS_EXCEPTION> is returned.
                             */
 );

/* @func
 *
 *
 * ===========================================================
 *
 *   Please use <f RfcIndirectCallEx> instead of this function
 *
 * ===========================================================
 *
 *
 * calls a function module in R/3 indirectly.
 *
 * With this function the call of a function module in R/3 will use the
 * transactional RFC interface in R/3.
 *
 * Importing parameters are not supported.
 *
 * If an error occurs (almost communication error) the RFC client program
 * has to reconnect to R/3 later and repeat this RFC call with the specific
 * TransID. It must NOT create a new TransID via <f RfcCreateTransID>.
 *
 * If the tRFC client program takes care of the TID management and repeats
 * this RFC call in error cases the function module will be executed
 * exactly once in R/3.
 *
 * @xref <f RfcCreateTransID>, <f RfcIndirectCallEx>
*/
RFC_RC SAP_API RfcIndirectCall(
 RFC_HANDLE       handle,    /* @parm connection handle        */
 rfc_char_t *           function,  /* @parm function module to call  */
 RFC_PARAMETER *  exporting, /* @parm 'exporting' parameters   */
 RFC_TABLE     *  tables,    /* @parm 'tables' parameters      */
 RFC_TID          tid        /* @parm <t RFC_TID>              */
 );

/* @func
 *
 * calls a function module in R/3 indirectly (using the tRFC interface
 * in R/3.
 *
 * Importing parameters are not supported.
 *
 * In error cases, the tRFC client
 * has to reconnect to R/3 later and repeat this RFC call with the specific
 * TransID. It must NOT create a new TransID via <f RfcCreateTransID>.
 *
 * After this call is returned without any error, the tRFC client must
 * issue <f RfcConfirmTransID> to confirm the executed function module.
 *
 * If the tRFC client program takes care of the TID management and repeats
 * this RFC call in error cases, the according function module will be executed
 * exactly once in R/3.
 *
 * Even if the tRFC client is broken down after this call is
 * successfully returned and before the update in the own TID management,
 * it can repeat this call with the same TID. The R/3 system will check
 * whether this transaction is already executed or not before the according
 * function module will be called.
 *
 * This is why we recommend you to use this new call instead of
 * <f RfcIndirectCall>.
 *
 * @xref <f RfcCreateTransID>
*/
RFC_RC SAP_API RfcIndirectCallEx(
 RFC_HANDLE       handle,    /* @parm connection handle        */
 rfc_char_t *           function,  /* @parm function module to call  */
 RFC_PARAMETER *  exporting, /* @parm 'exporting' parameters   */
 RFC_TABLE     *  tables,    /* @parm 'tables' parameters      */
 RFC_TID          tid        /* @parm <t RFC_TID>              */
 );


/* @func
 *
 * When using this function instead of <f RfcIndirectCallEx> the called
 * function module in R/3 won't be immediately executed but it will be inserted in
 * the receive queue from within R/3 defined by the parameter qname.
 *
 * A counter is used in order to execute the queue elements (function modules).
 *
 * R/3 will use the counter defined by the parameter qcount. But if qcount is 0
 * it will create a new counter for this transaction. The executing order is then
 * the order of elements being inserted in this queue.
 *
 * For every insert in this queue, we recommend you to use only the R/3 counter
 * or your own counter because the behavior of the mixed mode (sometimes with R/3
 * sometimes with your own counter) is not defined.
 *
 * When using your own counter you should use a specific queue
 * name and only your application will insert data in this queue.
 *
 * Using this functionality, the according queue must be activated by calling
 * the function module "TRFC_QIN_RESTART" with the parameter qname. You can call
 * this function module from within R/3 or from a remote system via RFC.
 *
 * Calling the function module "TRFC_QIN_RESTART" from a batch job, you can let
 * all according function modules in a receive queue run in a batch process and
 * you won't have problem with the max. ABAP-timeout in a dialog work process when
 * using RFC or tRFC with RfcIndirectCall or RfcIndirectCallEx.
 *
 * Importing parameters are not supported.
 *
 * In error cases, the tRFC client has to reconnect to R/3 later and repeat this
 * RFC call with the specific TransID. It must NOT create a new TransID via
 * <f RfcCreateTransID>.
 *
 * After this call is returned without any error, the tRFC client must
 * issue <f RfcConfirmTransID> to confirm the executed function module.
 *
 * If the tRFC client program takes care of the TID management and repeats
 * this RFC call in error cases, the according function module will be executed
 * exactly once in R/3.
 *
 * Even if the tRFC client is broken down after this call is
 * successfully returned and before the update in the own TID management,
 * it can repeat this call with the same TID. The R/3 system will check
 * whether this transaction is already executed or not before the according
 * function module will be called.
 *
 *
 * @xref <f RfcCreateTransID>
*/
RFC_RC SAP_API RfcQueueInsert(
 RFC_HANDLE       handle,    /* @parm connection handle        */
 rfc_char_t *     function,  /* @parm function module to call  */
 RFC_PARAMETER *  exporting, /* @parm 'exporting' parameters   */
 RFC_TABLE     *  tables,    /* @parm 'tables' parameters      */
 RFC_QNAME        qname,     /* @parm Queue name               */
 RFC_INT          qcount,    /* @parm Queue counter            */
 RFC_TID          tid        /* @parm <t RFC_TID>              */
 );


/* ---------------------------------------------
 * Receiving an RFC function call
 * -------------------------------------------*/

/* Reading the function name */

/*----------------------------------------------------------------------
@func
Besides using <f RfcDispatch> it is also possible
to receive RFC call directly. The function
RfcGetName must then be used to get the name of the called function.

If no error occurs, the returned function name will be a zero
terminated string.

The calling program then has to determine the interface of the
requested function module and to receive the parameters as within
a function being installed via <f RfcInstallFunction>.

In register mode (s. <f RfcAccept>) the server program will wait for
the next RFC requests.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

@xref <f RfcGetNameEx>, <f RfcDispatch>, <f RfcInstallFunction>, <f RfcGetData>, <f RfcSendData>
---------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetName (
  RFC_HANDLE   handle,           /* @parm RFC connection handle */
  RFC_FUNCTIONNAME functionname  /* @parm output parameter,
                                  * name of the called function module
                                  * ( <t RFC_FUNCTIONNAME> ).
                                  */
                          );

/*----------------------------------------------------------------------
@func
Using <f RfcGetName>, it is not possible
to receive an RFC function name longer that 30 chars.
RfcGetLongName must then be used to get the name of the called function if the length
of function name is longer than 30.

If no error occurs, the returned function name will be a zero
terminated string.

The calling program then has to determine the interface of the
requested function module and to receive the parameters as within
a function being installed via <f RfcInstallFunction>.

In register mode (s. <f RfcAccept>) the server program will wait for
the next RFC requests.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

@xref <f RfcGetLongNameEx>, <f RfcDispatch>, <f RfcInstallFunction>, <f RfcGetData>, <f RfcSendData>
---------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetLongName (
  RFC_HANDLE   handle,           /* @parm RFC connection handle */
  char*        functionname,     /* @parm output parameter,
                                  * name of the called function module
                                  * ( <t RFC_FUNCTIONNAME> ).
                                */
  unsigned      fn_length       /* @parm input parameter length of functionname */
                          );


/*----------------------------------------------------------------------
@func
Because the <f RfcGetName> is a blocking call, some RFC server programs
must loop with <f RfcListen> or <f RfcWaitForRequest> before calling
RfcGetName.

Till now, if an RFC system call such as RFC_PING for connection test
is incoming, it will be automatically executed by the RFC library.
After that, the RFC library does not return the control to the RFC program
but waits for the next RFC request.

This is necessary for RFC programs working with RfcGetName in "normal"-mode
(starting by an R/3 component) or in "register"-mode without RfcListen
or RfcWaitForRequest.

RFC programs running in "register"-mode and working with RfcListen/RfcWaitForRequest
can use this new function to have the control back after each RFC
system call. In this case, <e RFC_RC.RFC_SYSTEM_CALLED> will be returned
and the RFC program must check this code and repeat the call.

This new function is also a blocking call in "register"-mode.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE> or <e RFC_RC.RFC_SYSTEM_CALLED>.

@xref <f RfcGetName>, <f RfcDispatch>, <f RfcInstallFunction>, <f RfcGetData>, <f RfcSendData>
---------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetNameEx (
  RFC_HANDLE   handle,           /* @parm RFC connection handle */
  RFC_FUNCTIONNAME functionname  /* @parm output parameter,
                                  * name of the called function module
                                  * ( <t RFC_FUNCTIONNAME> ).
                                  */
                            );

/*----------------------------------------------------------------------
@func
Because the <f RfcGetLongName> is a blocking call, some RFC server programs
must loop with <f RfcListen> or <f RfcWaitForRequest> before calling
RfcGetName.

Till now, if an RFC system call such as RFC_PING for connection test
is incoming, it will be automatically executed by the RFC library.
After that, the RFC library does not return the control to the RFC program
but waits for the next RFC request.

This is necessary for RFC programs working with RfcGetlongName in "normal"-mode
(starting by an R/3 component) or in "register"-mode without RfcListen
or RfcWaitForRequest.

RFC programs running in "register"-mode and working with RfcListen/RfcWaitForRequest
can use this new function to have the control back after each RFC
system call. In this case, <e RFC_RC.RFC_SYSTEM_CALLED> will be returned
and the RFC program must check this code and repeat the call.

This new function is also a blocking call in "register"-mode.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE> or <e RFC_RC.RFC_SYSTEM_CALLED>.

@xref <f RfcLongGetName>, <f RfcDispatch>, <f RfcInstallFunction>, <f RfcGetData>, <f RfcSendData>
---------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetLongNameEx (
  RFC_HANDLE   handle,           /* @parm RFC connection handle */
  char*        functionname,     /* @parm output parameter,
                                  * name of the called function module
                                  * ( <t RFC_FUNCTIONNAME> ).
                                 */
 unsigned fn_length              /* @parm length of function name */
                            );





/* reading the import parameters */

/*---------------------------------------------------------------------
@func

Within a function registered via <f RfcInstallFunction>
or after receiving the name of the
called function by <f RfcGetName> the function RfcGetData can be
used to receive the parameters of the function call.

Here the <t ITAB_H> field in the <t RFC_TABLE> record has to be
initialized to NULL. The function RfcGetData fills in the
corresponding table handle, which is either a newly created table
or an already existing table, which was send to the caller via
another RFC call.  The field itmode in an <t RFC_TABLE> record
determines if a received table is passed by reference or by
value.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.
---------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetData (
  RFC_HANDLE   handle,          /* @parm RFC connection handle */
  RFC_PARAMETER *  parameters,  /* @parm 'importing' parameters
                                 * ( <t RFC_PARAMETER> ).
                                 */
  RFC_TABLE     *  tables );    /* @parm 'tables' parameters
                                 * ( <t RFC_TABLE> ).
                                 */





/*---------------------------------------------------------------------
 * @enum RFC_ONGETPARAMETER_ACTION |
 *
 *   Possibly return values for <f RFC_ONGETPARAMETER> function.
 *
 * @xref <f RFC_ONGETPARAMETER_ACTION>, <f RfcDynamicGetData>.
 *---------------------------------------------------------------------*/
typedef enum
{
 RFC_ONGETPARAMETER_ACTION_NOT_FOUND,        /* @emem any error has been occured */
 RFC_ONGETPARAMETER_ACTION_SINGLE_PARAMETER, /* @emem a single parameter have been received */
 RFC_ONGETPARAMETER_ACTION_SINGLE_TABLE,     /* @emem a single table have been received */
 RFC_ONGETPARAMETER_ACTION_ALL               /* @emem all parameters have been received */
} RFC_ONGETPARAMETER_ACTION;







/*------------------------------------------------------------------------
 * @cb RFC_ONGET_PARAMETER_ACTION |		RFC_ONGETPARAMETER |
 *
 * Prototype for callback function, pointer on which should be used as one of parameters
 * by <f RfcDynamicGetData>.
 *------------------------------------------------------------------------*/
typedef RFC_ONGETPARAMETER_ACTION ( DLL_CALL_BACK_FUNCTION_PTR RFC_ONGETPARAMETER )
        (
		  void           *context,             /* @parm pointer on a data area contains parameter */
          char           *parameterName,       /* @parm name of parameter, restored by <f RfcDynamicGetData> */
          unsigned        parameterNameLength, /* @parm name's length of parameter, restored by <f RfcDynamicGetData> */
          RFC_PARAMETER **parameters,          /* @parm pointer on parameters array */
          RFC_TABLE     **tables               /* @parm pointer on tables array */
		);







/*--------------------------------------------------------------------
 * @func
 *
 * A callback inorder to support filtering/preprocessing of parameters.
 * This API retrieves parameters one by one.
 *
 * When you first call this API you pass it a callback function. As the library is
 * unmarshaling parameters and tables, it calls the callback function with the
 * parameter name, and the callback function essentially returns an associated
 * RFC_PARAMETER or RFC_TABLE.
 *
 * @rdesc
 *
 * <e RFC_RC.RFC_OK> if call succesfull.
 *
 * <e RFC_RC.RFC_FAILURE> if call fails.
 *
 * @xref
 * <f RFC_ONGETPARAMETER>, <f RFC_ONGETPARAMETER_ACTION>.
 *
 * @ex Using RfcDynamicGetData |
 *
 *   ....
 *
 * typedef struct
 * {
 *  RFC_INT	sum1;
 *  RFC_INT	sum2;
 *
 *  RFC_PARAMETER  rp;
 *  RFC_TABLE	tables[2];
 *
 * } PT_CTX;
 *
 * static RFC_ONGETPARAMETER_ACTION onGetParameter
 *                                   (void*, char*, unsigned, RFC_PARAMETER**, RFC_TABLE**);
 *
 * static RFC_RC abc_function(RFC_HANDLE rfc_handle);
 *
 *  ....
 *
 * int main (int argc, char** argv)
 * {
 *  RFC_HANDLE 	rfc_handle;
 *  RFC_RC		rfc_rc;
 *  char*		docu;
 *
 *
 *
 *  printf ("Accept Rfc Connection ... ");
 *  rfc_handle = RfcAccept (argv);
 *  if (rfc_handle == RFC_HANDLE_NULL) rfc_error ("RfcAccept");
 *  printf ("Ok\n");
 *
 *
 *
 *  printf ("RfcInstallFunction ABCD ...");
 *  rfc_rc = RfcInstallFunction ("ABCD", abc_function, docu);
 *  if (rfc_rc != RFC_OK) rfc_error("Install Function ABC ");
 *  printf ("Ok\n");
 *
 *
 *  do
 *  {
 *  if (rfc_rc == RFC_OK) printf ("RfcDispatch called\n");
 *   rfc_rc = RfcDispatch (rfc_handle);
 *  } while (rfc_rc == RFC_OK);
 *
 *  RfcClose (rfc_handle);
 *
 *  printf ("Server Ok\n");
 *
 *  return 0;
 * }
 *
 * static RFC_RC abc_function(RFC_HANDLE rfc_handle)
 * {
 *  RFC_RC rfc_rc;
 *  RFC_PARAMETER  export[2];
 *  RFC_INT	result;
 *  PT_CTX		ctx;
 *
 *  int i;
 *
 *
 *  printf ("RfcDynamicGetData ... ");
 *  fflush(stdout);
 *  rfc_rc = RfcDynamicGetData (rfc_handle,
 *                      	     onGetParameter,
 *                       	     (void*) &ctx);
 *  if (rfc_rc != RFC_OK)
 *  {
 *   rfc_error ("RfcDynamicGetData");
 *   exit (1);
 *  }
 *  printf ("OK\n");
 *
 *
 *  printf ("Received import parameter sum1 has value:  %d\n", ctx.sum1);
 *  printf ("Received import parameter sum2 has value:  %d\n", ctx.sum2);
 *  printf ("Received table has name: %s \n", ctx.tables[0].name);
 *  printf ("Received table has content:\n");
 *  display_table (ctx.tables[0].ithandle);
 *
 *
 *  printf ("Assign to parameter result the value of (sum1 + sum2)\n");
 *  result = ctx.sum1 + ctx.sum2;
 *
 *  printf ("Insert new entries into table %s ... ", ctx.tables[0].name);
 *  update_table(ctx.tables[0].ithandle);
 *  printf ("Ok\n");
 *
 *  ctx.tables[1].name = NULL;
 *
 *  export[0].name  = "RESULT";
 *  export[0].nlen  = strlen (export[0].name);
 *  export[0].type  = RFCTYPE_INT;
 *  export[0].addr  = &result;
 *  export[0].leng  = sizeof (result);
 *
 *  export[1].name = NULL;
 *
 *
 *  printf ("RfcSendData ... ");
 *  rfc_rc = RfcSendData (rfc_handle,
 * 		       export,
 * 		       ctx.tables);
 *  if (rfc_rc != RFC_OK)
 *  {
 *   rfc_error ("RfcSendData");
 *   exit (1);
 *  }
 *  printf ("Ok\n");
 *
 *
 *  return RFC_OK;
 * }
 *
 *
 * static RFC_ONGETPARAMETER_ACTION onGetParameter
 * ( void* context,
 *   char* parameterName,
 *   unsigned leng,
 *   RFC_PARAMETER** parameters,
 *   RFC_TABLE** tables
 * )
 * {
 *  PT_CTX 	*pCTX;
 *  RFC_PARAMETER	**ppParam;
 *  RFC_TABLE      **ppTable;
 *
 *  pCTX = (PT_CTX*) context;
 *
 *  if ( (memcmp ("sum1", parameterName, leng) == 0) || (memcmp ("SUM1", parameterName, leng) == 0) )
 *  {
 *   *parameters = &(pCTX->rp);
 *   *tables = NULL;
 *
 *   pCTX -> rp.name = "sum1";
 *   pCTX -> rp.nlen = 4;
 *   pCTX -> rp.type = TYPINT;
 *   pCTX -> rp.addr = &(pCTX->sum1);
 *   pCTX -> rp.leng = sizeof (TYPINT);
 *
 *   return RFC_ONGETPARAMETER_ACTION_SINGLE_PARAMETER;
 *  }
 *
 *  if ( (memcmp ("sum2", parameterName, leng) == 0) || (memcmp ("SUM2", parameterName, leng) == 0) )
 *  {
 *   *parameters = &(pCTX->rp);
 *   *tables = NULL;
 *
 *   pCTX -> rp.name = "sum2";
 *   pCTX -> rp.nlen = 5;
 *   pCTX -> rp.type = TYPINT;
 *   pCTX -> rp.addr = &(pCTX->sum2);
 *   pCTX -> rp.leng = sizeof (TYPINT);
 *
 *   return RFC_ONGETPARAMETER_ACTION_SINGLE_PARAMETER;
 *  }
 *
 *  if (memcmp ("ITAB1000", parameterName, leng) == 0)
 *  {
 *   *tables = pCTX->tables;
 *   *parameters = NULL;
 *
 *   pCTX -> tables[0].name = "ITAB1000";
 *   pCTX -> tables[0].nlen = 8;
 *   pCTX -> tables[0].type = TYPC;
 *   pCTX -> tables[0].leng = 1000;
 *   pCTX -> tables[0].ithandle = ITAB_NULL;
 *   pCTX -> tables[0].itmode = RFC_ITMODE_BYREFERENCE;
 *
 *   return RFC_ONGETPARAMETER_ACTION_SINGLE_TABLE;
 *  }
 *
 *  return RFC_ONGETPARAMETER_ACTION_NOT_FOUND;
 *
 * }
 *
 *
 *----------------------------------------------------------------------*/
RFC_RC SAP_API RfcDynamicGetData
(
 RFC_HANDLE           handle,         /* @parm RFC connection handle. */
 RFC_ONGETPARAMETER   onGetParameter, /* @parm pointer on user implemented
									   * function which restores the parameter
									   * according their names.
									   */
 void                *context		  /* @parm data area containing parameter have to be received. */
);













/*------------------------------------------------------------------------
@func
This function is used to send back the answer of a Remote Function Call
to a caller .

The tables description ( <t RFC_TABLE> ) must be the same than in the
previous RfcGetData call.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.
------------------------------------------------------------------------*/
RFC_RC SAP_API RfcSendData(
  RFC_HANDLE   handle,        /* @parm RFC connection handle */
  RFC_PARAMETER *  parameters, /* @parm 'exporting' parameters
                                */
  RFC_TABLE     *  tables );   /* @parm 'tables' parameters.
                                * Must be the same structure as passed to
                                * <f RfcGetData>.
                                */

/* raise an exception */
/*------------------------------------------------------------------------
@func
This function is used to raise an exception while processing a
Remote Function Call.

If you are having tables parameters in your function module, use
<f RfcRaiseTables> instead.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

@xref
<f RfcRaiseTables>
------------------------------------------------------------------------*/
RFC_RC SAP_API RfcRaise(
 RFC_HANDLE      handle,     /* @parm RFC connection handle */
 rfc_char_t *    exception   /* @parm exception to be raised
                              * (zero terminated string).
                              */
                       );
/* raise an exception (with tables parameters) */
/*------------------------------------------------------------------------
@func
This function is used to raise an exception while processing a
Remote Function Call, if the function module being called has
tables parameters.

@rdesc
<e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.

@xref
<f RfcRaise>
------------------------------------------------------------------------*/
RFC_RC SAP_API RfcRaiseTables(
 RFC_HANDLE      handle,     /* @parm RFC connection handle */
 rfc_char_t *    exception, /* @parm exception to be raised
                              * (zero terminated string).
                              */
 RFC_TABLE  *    tables     /* @parm tables parameters as received by
                              * <f RfcGetData>.
                              */
                       );


/*------------------------------------------------------------------------
 * @func
 * This function is used to raise an exception and send back an error message. 
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>.
 *
 * @xref
 * <f RfcRaiseTables>
 *------------------------------------------------------------------------*/
RFC_RC SAP_API RfcRaiseErrorMessage
(
 RFC_HANDLE      handle,	/* @parm RFC connection handle */
 rfc_char_t *    exception, /* @parm exception to be raised
							 * (zero terminated string).
                             */
 RFC_TABLE  *    tables,    /* @parm tables parameters as received by
                             * <f RfcGetData>.
                             */
 rfc_char_t * szErrorMsg    /* @parm Error message to be sent to the caller. The
                             * maximum length of the message is 200.
							 */                         

);


/*-----------------------------------------------------------------------------
@func

can be used to check if there is a RFC request available.

If one does not always want to wait for the answer to an RFC call
by <f RfcReceive>  one has to call the function RfcListen
to listen for incoming RFC events.

The function always returns immediately.
If RfcListen returns <e RFC_RC.RFC_OK>, RfcReceive has to be called to
handle the event. It only is possible  to listen for one incoming
RFC message at a time.

It is also possible to use RfcListen while waiting for a new RFC request
by <f RfcDispatch> with or without register mode (s. <f RfcAccept>). Because
RfcListen will immediately return, it is recommended to use <f RfcWaitForRequest>
to wait for new RFC requests.

@rdesc
The function returns  <e RFC_RC.RFC_OK>,
if there is an RFC event pending (call or return) and  <e RFC_RC.RFC_RETRY>,
if nothing has arrived yet. The function returns <e RFC_RC.RFC_FAILURE>, if
an error has occurred .

@ex Using RfcListen while waiting for the answer of a remote function call. |

RFC_RC rc;

rc = RfcCall( handle, .... );

do
{
   rc = RfcListen( handle );
   if( rc == RFC_RETRY )
   {
      // while waiting for the RFC answer, do something...
      ...
   }
} while ( rc == RFC_RETRY );

if( rc == RFC_OK )
   rc = RfcReceive( handle, .... );
...

@ex Using RfcListen and RfcDispatch. |

RFC_RC     rc;
RFC_HANDLE handle = RfcAccept( argv );
...
do
{
  // Waiting for the next RFC request
  for( rc = RFC_RETRY; rc == RFC_RETRY; )
  {
    rc = RfcListen( handle );
    if( rc == RFC_RETRY )
    {
      // while waiting for the next RFC request, do something...
      ....
    }
  }
  if (rc != RFC_OK)
    break;

  rc = RfcDispatch( handle );

} while( rc == RFC_OK );

RfcClose( handle );
exit( 0 );


-----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcListen(
  RFC_HANDLE handle  /* @parm RFC connection handle */
  );

/*-----------------------------------------------------------------------------
@func

After <f RfcAccept> using register mode, this function can be used to wait
for a defined time for new RFC requests.

The time must be defined in sec (-1: infinit).

If the RFC server program doesn't run in register mode, this call has
the same functionality as RfcListen (immediately returning).

-----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcWaitForRequest(
  RFC_HANDLE handle, /* @parm RFC connection handle */
  RFC_INT    wtime   /* @parm Wait time             */
  );

RFC_RC SAP_API RfcFileArgv(int* argc, rfc_char_t*** PntArgv);

/*---------------------------------------------------------------------------
@func

This function can be used to build up the structures <t RFC_OPTIONS>,
<t RFC_CONNOPT_R3ONLY>, <t RFC_CONNOPT_VERSION_3> or
<t RFC_CONNOPT_CPIC> from the command line.

@comm

The following tokens are recognized in the argv array, which must
be terminated by a NULL entry :

@flag -d <lt>Destination<gt>  | the name of the destination
@flag -c <lt>NNN<gt>          | client (sign on data)
@flag -u <lt>User Id<gt>      | user id
@flag -p <lt>Password<gt>     | password
@flag -l <lt>language<gt>     | language
@flag -3                      | R/3 mode
@flag -2                      | CPIC mode
@flag -t                      | turn trace on
@flag -h <lt>Hostname<gt>        | the name of the target host
@flag -s <lt>NN<gt>              | the system number of the target SAP system
@flag -g <lt>Gateway Host<gt>    | the gateway host (if not specified, the
                          -h option is used)
@flag -x <lt>Gateway Service<gt> | the TCP/IP service of the gateway (
                             default is sapgwNN, where NN is
                             the system number (-s)
@flag -gui | start 'sapgui' to be able to display SAP Dynpros or
             Graphics (R/3 mode only).
@flag -debug | start communication in debug mode (R/3 mode only).

All tokens that were interpreted by RfcConnArgv3 are removed from
the argv array.

@rdesc returns 0.

@xref <f RfcConnArgv>

---------------------------------------------------------------------------*/
int    SAP_API RfcConnArgv3(
 rfc_char_t **                 argv,         /* @parm command line to be parsed. */
 RFC_OPTIONS *           rfc_opt,      /* @parm option structure to
                                        * be filled.
                                        */
 RFC_CONNOPT_CPIC *      connopt_cpic, /* option extension for R/2 connection
                                        * (a valid pointer must be supplied
                                        *  even if you are not using a R/2
                                        *  system).
                                        */
 RFC_CONNOPT_R3ONLY *    connopt_r3only,  /* @parm option extension for R/3 connection
                                           * (a valid pointer must be supplied
                                           *  all the time).
                                           */
 RFC_CONNOPT_VERSION_3 * connopt_version_3 /* @parm option extension for a
                                            * connection
                                            * to a R/3 system of Rel. 3.0
                                            * or later
                                            * (a valid pointer must be supplied
                                            *  all the time).
                                            */
                           );

/*---------------------------------------------------------------------------
@func

This function was used to build up the structures <t RFC_OPTIONS> ,
<t RFC_CONNOPT_R3ONLY> or
<t RFC_CONNOPT_CPIC> from the command line in former releases and
is included for backward compatibility.

It offers the same functionality as <f RfcConnArgv3> allocating
a static instance of <t RFC_CONNOPT_VERSION_3> internally, however.

Use the function <f RfcConnArgv3> instead, if you are not working
on 16-bit Windows.

On 16-bit Windows only RfcConnArgv is supported.

@rdesc returns 0.

@xref <f RfcConnArgv3>

---------------------------------------------------------------------------*/

int    SAP_API RfcConnArgv(
 rfc_char_t **                 argv,         /* @parm command line to be parsed. */
 RFC_OPTIONS *           rfc_opt,      /* @parm option structure to be filled. */
 RFC_CONNOPT_CPIC *      connopt_cpic, /* option extension for R/2 connection
                                        * (a valid pointer must be supplied
                                        *  even if you are not using a R/2
                                        *  system).
                                        */
 RFC_CONNOPT_R3ONLY *    connopt_r3only   /* @parm option extension for R/3 connection
                                           * (a valid pointer must be supplied
                                           *  all the time).
                                           */
                           );

RFC_RC SAP_API RfcSendSystemInfo( RFC_HANDLE handle );
RFC_RC SAP_API RfcSendPing      ( RFC_HANDLE handle );
RFC_RC SAP_API RfcSendDocu      ( RFC_HANDLE handle );


/* @cb  RFC_RC |  RFC_ONCALL |
 *
 * type of C function to be installed by <f RfcInstallFunction>
 * or <f RfcInstallFunctionExt>.
 *
 */
typedef RFC_RC ( DLL_CALL_BACK_FUNCTION_PTR RFC_ONCALL ) (
   RFC_HANDLE handle /* @parm Handle to RFC connection, from where the RFC parameters
                      * must be read and to which the parameters must be returned.
                      */
               );

/*--------------------------------------------------------------------------
@func

Used to install a function to be callable as RFC function module.

For receiving an RFC call there are two possible ways. The most
simple way to receive an RFC call in an external program is to
register a C function to be called if a call request is received.

The function RfcInstallFunction registers a C function to be
called when receiving the request for an RFC call.

The function pointer points to a function of type <f RFC_ONCALL>,
which contains the functionality being offered as an RFC function
module. 'functionname' is the name of the offered RFC function
module and description should contain a description of  the
functionality as well as a description of the interface. Newline
characters can be used to start new lines.

The descriptions of the registered functions can be requested by
calling the standard function module RFC_DOCU being available in
every RFC server program using <f RfcDispatch> interface.

After <f RfcAccept> or after receiving the return code <e RFC_RC.RFC_CALL>,
when calling <f RfcReceive> the program has to call <f RfcDispatch>,
which internally calls the corresponding registered function.

@rdesc
returns <e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>, if there is no memory available
to register the function.

@xref <f RfcDispatch>, <f RfcGetName>

@comm
This function is not available for 16-Bit Windows (Windows 3.x).
Use <f RfcInstallFunctionExt> instead.
 --------------------------------------------------------------------------*/
RFC_RC SAP_API RfcInstallFunction(
 RFC_FUNCTIONNAME functionname,  /* @parm name of function as it can be called
                                  * from ABAP/4 environment. Zero terminated
                                  * string.
                                  */
 RFC_ONCALL       f_ptr,         /* @parm function to be called. Must be of type
                                  * <f RFC_ONCALL>.
                                  */
 rfc_char_t *           docu           /* @parm text describing the functionality and
                                  * the parameters of the function module.
                                  */
                                 );

/*-------------------------------------------------------------------------------
@func

Alternate function replacing <f RfcInstallFunction> for Windows 3.x (16-Bit).
The functions needs a valid RFC handle as an additional parameter.

The function module is installed only for the given RFC connection.

@rdesc
returns <e RFC_RC.RFC_OK> or <e RFC_RC.RFC_FAILURE>, if there is no memory available
to register the function.

@xref <f RfcInstallFunction>, <f RfcDispatch>, <f RfcGetName>

@comm
The function is available on all supported platforms to ease porting.
-----------------------------------------------------------------------------------*/
RFC_RC SAP_API RfcInstallFunctionExt(
 RFC_HANDLE handle,              /* @parm valid RFC connection handle,
                                  * as returned by <f RfcAccept>.
                                  */
 RFC_FUNCTIONNAME functionname,  /* @parm name of function as it can be called
                                  * from ABAP/4 environment. Zero terminated
                                  * string.
                                  */
 RFC_ONCALL       f_ptr,         /* @parm function to be called. Must be of type
                                  * <f RFC_ONCALL>.
                                  */
 rfc_char_t *           docu           /* @parm text describing the functionality and
                                  * the parameters of the function module.
                                  */

  );

/* @func
 * function to be called to process the next RFC request.
 *
 * After <f RfcAccept> or after receiving the return code
 * <e RFC_RC.RFC_CALL>, when calling <f RfcReceive> or
 * <f RfcCallReceive> the program has to call this function
 * to process the pending RFC request.
 *
 * This function internally calls
 * the registered function corresponding to the RFC call.
 * The return code of the registered function is again
 * returned by RfcDispatch.
 *
 * @comm
 * There are some function modules, which are always
 * offered automatically, when using RfcDispatch.
 *
 * These are
 * @flag RFC_DOCU | get of list of the installed function
 *                  modules.
 * @flag RFC_PING | do nothing, used for connection test.
 * @flag RFC_SYSTEM_INFO | deliver some information on
 *                         the RFC library used.
 *
 * @rdesc
 * returns <e RFC_RC.RFC_OK> or other <t RFC_RC> return code.
 *
 * @ex a typical RFC server |
 *
 * int main ( int argc, char ** argv )
 * {
 *   RFC_HANDLE handle;
 *   RFC_RC     rc;
 *
 *   handle = RfcAccept( argv );
 *   if ( handle == RFC_HANDLE_NULL )
 *   {
 *      ... error processing
 *      return 1;
 *   }
 *
 *   rc = RfcInstallFunction( ..... );
 *   if( rc != RFC_OK )
 *   {
 *      ... error processing
 *      return 1;
 *   }
 *
 *   do
 *   {
 *      rc = RfcDispatch( handle );
 *   }
 *   while( rc == RFC_OK );
 *
 *   RfcClose( handle );
 *
 *   return 0;
 * }
 *
 */
RFC_RC SAP_API RfcDispatch(
  RFC_HANDLE  handle   /* @parm handle to RFC connection */
   );

/* @cb  int |  RFC_ON_CHECK_TID |
 *
 * function to be installed by <f RfcInstallTransactionControl>.
 *
 * The function is called when a local transaction is starting.
 * Since a transactional RFC call can be issued many times
 * by the client system, the function is responsible for storing
 * the transaction ID in permanent storage.
 * If the client system
 * tries starting the same transaction
 * a second time, the function has to return 1.
 *
 * @comm
 * If this functions has access to a SQL database,
 * it should perform a operation like
 *
 *    Insert Into SomeTable values ( :transactionId );
 *
 * where the table 'SomeTable' must have a unique index
 * over the transaction ID. If another process has also inserted
 * the same transaction ID, the second process has to wait until
 * the transaction is completed by the first process. If the
 * transaction is completed by another process (The Insert command
 * returns with some 'Duplicate Record' error code, the function must
 * return 1 to indicate to skip the transaction.
 *
 * @rdesc
 * return values
 * @flag 0      | transaction ID stored, transaction can be started.
 * @flag 1      | transaction already done, skip the request.
 * @flag <lt> 0 | cannot lock transaction, internal error.
 *
 *
 */
typedef int ( DLL_CALL_BACK_FUNCTION_PTR RFC_ON_CHECK_TID )(
  RFC_TID transactionId  /* @parm <t RFC_TID> */
   );

/* @cb  int |  RFC_ON_COMMIT |
 *
 * function to be installed by <f RfcInstallTransactionControl>.
 *
 * The function is called when a local transaction ends.
 *
 * The function is to be used to commit the local transaction,
 * if necessary.
 *
 * @comm
 * If this functions has access to a SQL database,
 * it should perform a operation like
 *
 *    Commit work;
 *
 */
typedef void ( DLL_CALL_BACK_FUNCTION_PTR RFC_ON_COMMIT )
(
    RFC_TID transactionId  /* @parm <t RFC_TID> */
);


/* @cb  int |  RFC_ON_ROLLBACK |
 *
 * function to be installed by <f RfcInstallTransactionControl>.
 *
 * The function is called when a local transaction ends with failure.
 *
 * The function is to be used to roll back the local transaction,
 * if necessary.
 *
 * @comm
 * If this functions has access to a SQL database,
 * it should perform a operation like
 *
 *    Rollback work;
 *
 */
typedef void ( DLL_CALL_BACK_FUNCTION_PTR RFC_ON_ROLLBACK )
(
    RFC_TID transactionId  /* @parm <t RFC_TID> */
);


/* @cb  void |  RFC_ON_CONFIRM_TID |
 *
 * function to be installed by <f RfcInstallTransactionControl>.
 *
 * The function is called when a local transaction is completed.
 * All information stored about that transaction can be discarded
 * by the server.
 *
 * In general this function can be used to delete the transaction
 * ID from permanent storage.
 *
 * @comm
 * If this functions has access to a SQL database,
 * it should perform a operation like
 *
 *    Delete From SomeTable where key = :transactionId; Commit Work;
 *
 * where the table 'SomeTable' should have a unique index
 * over the transaction ID.
 *
 *
 *
 */
typedef void ( DLL_CALL_BACK_FUNCTION_PTR RFC_ON_CONFIRM_TID )
(
    RFC_TID transactionId  /* @parm <t RFC_TID> */
);

/* @func
 *
 * installs 4 functions to control transactional behavior.
 * Only to be used with transactional RFC.
 *
 * Must be called by
 * RFC server program before the RfcDispatch loop is entered,
 * if this program wants to receive transactional RFC calls,
 * i.e. calls being issued by 'Call Function ... In Background Task'
 * and must ensure, that RFC calls are done exactly once.
 *
 * Without installing these functions it can only be guaranteed,
 * that an RFC function call issued by
 * 'Call Function ... In Background Task'
 * is done at least once. Then all function modules offered by such a server
 * program, which are called via 'Call Function ... In Background Task'
 * must cope with being called more then once.
 *
 * If installed, the first function onCheckTid is called, if a transactional
 * RFC is to be called. The actual Transaction ID
 * is passed ( <f RFC_ON_CHECK_TID> ).
 * The function has to store this transaction id in permanent
 * storage and return 0 .
 * If the same function will be called later again with the same
 * Transaction Id, it has to make sure that it will
 * return a non-zero value.
 * If the same transaction is already
 * running by another process, but is not completed,
 * the function has to wait, until
 * the transaction completes.
 *
 * The second function is called, if all the RFC function module
 * calls are done and the local transaction can be completed.
 * It should be used, to locally commit the transaction, if necessary
 * ( <f RFC_ON_COMMIT> ).  This function is 'void' .
 *
 * The third function is called instead of the 2nd function,
 * if, from the point of view of the RFC library,
 * there occurred an error while processing the local transaction.
 * This function can be used to roll back the local transaction
 * ( <f RFC_ON_ROLLBACK> ). This function is 'void' .
 *
 * The 4th function will be called, if the local transaction is
 * completed also from the point of view of the calling system
 * and all information on this transaction Id can be
 * discarded. This function is 'void' .
 * <f RFC_ON_CONFIRM_TID>
 *
 * @xref <t RFC_TID>
 */
void SAP_API RfcInstallTransactionControl (
 RFC_ON_CHECK_TID   onCheckTid, /* @parm function to be called, when
                                 * local transaction starts. Must be used
                                 * to check, if the transaction is already
                                 * running or was already completed.
                                 */
 RFC_ON_COMMIT      onCommit,   /* @parm function to be called, when
                                 * local transaction ends.
                                 * Should be used to commit the
                                 * transaction locally.
                                 */
 RFC_ON_ROLLBACK    onRollback, /* @parm function to be called, if
                                 * local transaction fails due to an
                                 * error found while the processing
                                 * is done inside the RFC library.
                                 * Should be used to roll back
                                 * a local transaction.
                                 */
 RFC_ON_CONFIRM_TID onConfirmTid /* @parm function to be called, when
                                  * local transaction is confirmed.
                                  * All information stored locally about
                                  * this transaction can be deleted.
                                  */
                                        );

/* SAP->API: info transact. rfc */

RFC_RC SAP_API RfcGetTrfcInfo( rfc_char_t * user, rfc_char_t * dest );


/* SAP->API: get transactionID and set f_ptr for confirmation */

RFC_RC SAP_API RfcGetTransID( RFC_RC (* _SYSTEM f_ptr)( rfc_char_t * ),
                                           RFC_TID tid );

/* @func
 *
 * Gets a Transaction-ID for a following call of a function module in R/3
 * using the transactional RFC interface in R/3
 *
 * With this function a new TransID will be produced from the R/3 system
 * the RFC client program has to call a function module in R/3 via
 * <f RfcIndirectCall> or <f RfcIndirectCallEx> with this TransID.
 *
 * If an error occurs (e.g. communication error) during the call of a function
 * module in R/3 via RfcIndirectCall the RFC client program has to reconnect
 * the RFC connection and repeat the RfcIndirectCall/RfcIndirectCallEx without
 * creating a new TransID.
 *
 * @xref <f RfcIndirectCall>, <f RfcIndirectCallEx>
 *
 */
RFC_RC SAP_API RfcCreateTransID(
 RFC_HANDLE       handle,    /* @parm connection handle. */
 RFC_TID          tid        /* @parm <t RFC_TID>        */
 );
/* API->SAP: check transID already is enqueued                */

RFC_RC SAP_API RfcCheckTransID(   RFC_HANDLE  handle,
                                  RFC_TID     tid       );

/* @func
 *
 * explizitely confirms a transaction using by transactional RFC.
 *
 * This call is only for use with <f RfcIndirectCallEx> and NOT
 * with <f RfcIndirectCall>.
 *
 * The function module called by <f RfcIndirectCallEx> is already
 * executed. This call is only necessary to delete some entries
 * for this TransID in SAP database. A repeat of this call in error
 * cases (almost communication error) is also available.
 *
 * If you don't use this call the SAP database will grow up.
 *
 * @xref <f RfcIndirectCallEx>
 *
 */
RFC_RC SAP_API RfcConfirmTransID(
 RFC_HANDLE       handle,    /* @parm connection handle. */
 RFC_TID          tid        /* @parm <t RFC_TID>        */
 );

/* @func
 *
 * explicitly allows the RFC library to start the programs described
 * in the list as input parameter if this is required by the RFC partner.
 *
 * This call can be used either in RFC client or RFC server program.
 *
 */
RFC_RC SAP_API RfcAllowStartProgram(
 rfc_char_t           * pname_list /* @parm zero terminated string.
			      * This is a list of programs which can
			      * be started by the RFC library by the
			      * RFC partner if required. A program
			      * name can have max. 512 B. They must
			      * be separated by ;.
			      * NULL: any program is allowed.
			      */
 );

/* @func
 *
 * activates the RFC trace for a specified connection or for all RFC
 * connections (RFC_HANDLE_NULL as handle).
 *
 * This call can be used either in RFC client or RFC server program.
 *
 */
void SAP_API RfcSetTrace(
 RFC_HANDLE       handle,      /* @parm connection handle
				* or RFC_HANDLE_NULL
				*/
 int              trace_level  /* @parm trace_level
				*  0: deactivate the RFC trace
				*  1: activate the RFC trace
				*/
 );


/* Internal use, not available in librfc */

RFC_HANDLE RfcLocalOpen( rfc_char_t * name );
RFC_HANDLE RfcLocalOpenAsync( void );
RFC_RC SAP_API RfcCallAsync( RFC_HANDLE      handle,
                             rfc_char_t *          function,
                             RFC_PARAMETER * parameters,
                             RFC_TABLE     * tables );
RFC_HANDLE RfcGetExternalDestination( rfc_char_t * name );

/*
    New API functions that omit the use of structs.
    d002076, 28.1.94
*/
typedef struct
{
        RFC_INT MaxNo;
    void* Params;
}RFC_PARAMETER_STACK;

typedef struct
{
    RFC_PARAMETER_STACK Exports;
    RFC_PARAMETER_STACK Imports;
    RFC_PARAMETER_STACK Tables;
} RFC_PARAMETER_ADMIN;

typedef RFC_PARAMETER_ADMIN* RFC_PARAM_SPACE;

RFC_PARAM_SPACE SAP_API RfcAllocParamSpace(unsigned MaxEx,
                                 unsigned MaxIm,
                                 unsigned MaxTab);

RFC_RC SAP_API RfcFreeParamSpace(RFC_PARAM_SPACE PSpace);

RFC_RC SAP_API RfcAddExportParam(RFC_PARAM_SPACE PSpace,
                                 unsigned ParamNo,
                                 void* name,
                                 unsigned nlen,
                                 unsigned type,
                                 unsigned leng,
                                 void* addr);

RFC_RC SAP_API RfcAddImportParam(RFC_PARAM_SPACE PSpace,
                                 unsigned ParamNo,
                                 void* name,
                                 unsigned nlen,
                                 unsigned type,
                                 unsigned leng,
                                 void* addr);

/* @func
 *
 * Special call for Visual Basic 4.0/5.0.
 * Instead of RfcAddImportParam() you must use RfcDefineImportParam()
 * to define all IMPORTING-parameters.
 * After RfcCallReceiveExt() or RfcReceiveExt() or RfcGetDataExt()
 * you have to issue RfcGetImportParam() for each IMPORTING-parameter
 * to get the contents of this parameter.
 *
 * @xref <f RfcGetImportParam>
 *
 */
RFC_RC SAP_API RfcDefineImportParam(
  RFC_PARAM_SPACE   PSpace,  /* @parm Pointer returned by RfcAllocParam */
  unsigned          ParamNo, /* @parm Parameter number (0, 1, 2, ...)   */
  void            * name,    /* @parm Parameter name                    */
  unsigned          nlen,    /* @parm Length of parameter name          */
  unsigned          type,    /* @parm Parameter type                    */
  unsigned          leng     /* @parm Length of parameter value         */
  );

/* @func
 *
 * Special call for Visual Basic 4.0/5.0.
 * Instead of RfcAddImportParam() you must use RfcDefineImportParam()
 * to define all IMPORTING-parameters. After RfcCallReceiveExt()
 * or RfcReceiveExt() or RfcGetDataExt() you have to issue
 * RfcGetImportParam() for each IMPORTING-parameter to get the
 * contents of this parameter.
 *
 * @xref <f RfcDefineImportParam>
 *
 */
RFC_RC SAP_API RfcGetImportParam(
  RFC_PARAM_SPACE   PSpace,  /* @parm Pointer returned by RfcAllocParam */
  unsigned          ParamNo, /* @parm Parameter number (0, 1, 2, ...)   */
  void            * addr     /* @parm Addr. for parameter value         */
  );

RFC_RC SAP_API RfcAddTable ( RFC_PARAM_SPACE PSpace,
                                 unsigned TableNo,
                                 void* name,
                                 unsigned nlen,
                                 unsigned type,
                                 unsigned leng,
                                 ITAB_H ithandle);

ITAB_H SAP_API RfcGetTableHandle(RFC_PARAM_SPACE PSpace,
                                 unsigned TableNo);

RFC_RC SAP_API RfcCallExt ( RFC_HANDLE handle,
			    RFC_PARAM_SPACE PSpace,
			    rfc_char_t* function);

RFC_RC SAP_API RfcReceiveExt(RFC_HANDLE handle,
			     RFC_PARAM_SPACE PSpace,
			     rfc_char_t* exception);

RFC_RC SAP_API RfcCallReceiveExt(RFC_HANDLE handle,
                                 RFC_PARAM_SPACE PSpace,
                                 rfc_char_t* function,
                                 rfc_char_t* exception);

RFC_RC SAP_API RfcIndirectCallExt(RFC_HANDLE handle,
                                  RFC_PARAM_SPACE PSpace,
                                  rfc_char_t* function,
                                  RFC_TID tid);

RFC_RC SAP_API RfcIndirectCallExExt(RFC_HANDLE handle,
                                    RFC_PARAM_SPACE PSpace,
                                    rfc_char_t* function,
                                    RFC_TID tid);

RFC_RC SAP_API RfcGetDataExt(RFC_HANDLE handle,
                                 RFC_PARAM_SPACE PSpace);

RFC_RC SAP_API RfcSendDataExt(RFC_HANDLE handle,
                                 RFC_PARAM_SPACE PSpace);

int SAP_API RfcExecProgram( rfc_char_t * c,
                            rfc_char_t * et,
                            unsigned leng );

RFC_INT SAP_API RfcGetVersion(void);
RFC_RC SAP_API RfcCheckVersionsOfUsedLibs(void);
RFC_RC SAP_API RfcGetAllLibVersions(rfc_char_t* buffer, int BufferLength);
RFC_RC SAP_API RfcStructureToFile( RFC_HANDLE handle,
				   rfc_char_t * tableName,
				   void * file /* must be FILE * ! */);

RFC_RC SAP_API RfcStructureToFileByName( RFC_HANDLE handle,
				         rfc_char_t * tableName,
				         rfc_char_t * filename );

/* @func
 *
 * The RFC library supports 2 functions to convert encoded decimal
 * numbers (BCD number) as defined in SAP systems (Type P) to a
 * character string and vice versa RfcConvertCharToBcd and
 * RfcConvertBcdToChar.
 *
 * A BCD number is a 16 byte structure which contains 31 digits and
 * the sign. The structure is aligned in a long-word boundary.
 *
 * The sign can have the values '+' or '-'. Analogous to SAP system,
 * the hex value 'C' and 'D' are used as positive and negative sign.
 *
 * This function converts a character string to a BCD number.
 *
 * @rdesc
 * return values
 * @flag 0      | OK
 * @flag 2      | "out_len" too small
 * @flag 3      | Invalid character in "in"
 * @flag 4      | Invalid "in_len"
 *
 * @xref <f RfcConvertBcdToChar>
 *
 */
int SAP_API RfcConvertCharToBcd(
  RFC_CHAR * in,       /* @parm Input data to convert */
  int        in_len,   /* @parm Length of input data  */
  int      * pdecs,    /* @parm No. of digits after , */
  RFC_BCD  * out,      /* @parm Output data           */
  int        out_len   /* @parm Length of output data */
  );

/* @func
 *
 * converts a BCD number into a character string.
 *
 * See <f RfcConvertCharToBcd> for more details.
 *
 *
 * @rdesc
 * return values
 * @flag 0      | OK
 * @flag 2      | "out_len" too small
 * @flag 3      | Invalid character in "in"
 * @flag 4      | Invalid "in_len"
 *
 */
int SAP_API RfcConvertBcdToChar(
  RFC_BCD  * in,       /* @parm Input data to convert */
  int        in_len,   /* @parm Length of input data  */
  int        decs,     /* @parm No. of digits after , */
  RFC_CHAR * out,      /* @parm Output data           */
  int        out_len   /* @parm Length of output data */
  );

/* 2-PC-RFC Identifier (TXID) */
#define RFC_SAP_TXID_LN           128
#define RFC_EXT_TXID_LN           32

/* @type RFC_SAP_TXID |
 * zero terminated string giving a 2-Phaser-Commit-RFC
 * globally unique identifier.
 *
 * ATTENTION:  Don't mix up this ID with the <t RFC_TID> using
 *             in transactional RFC.
 *
 */
typedef rfc_char_t RFC_SAP_TXID[RFC_SAP_TXID_LN+1];

/* @func
 *
 * RFC with 2-PHASE-COMMIT (2-PC-RFC)
 *
 * Only available with R/3 from 4.0C onwards and the called function
 * module in R/3 must use the Update Task for database updating.
 *
 *
 * The RFC library supports 4 functions for working with 2-PC-RFC:
 *
 * RfcTxStart             Start a transaction
 *
 * RfcTxPrepareToCommit   Prepare to commit the current transaction
 *
 * RfcTxCommit            Commit a transaction
 *
 * RfcTxAbort             Abort a transaction
 *
 *
 * @xref <f RfcTxPrepareToCommit>, <f RfcTxCommit>, <f RfcTxAbort>
 *
 */
RFC_RC SAP_API RfcTxStart(
 RFC_HANDLE          handle,       /* @parm connection handle.       */
 void *              ext_txid,     /* @parm External Transaction ID. */
 int                 ext_txid_len, /* @parm Length of the External
		   	            * Transaction ID, Max. 32 B.
			   	    */
 RFC_SAP_TXID        sap_txid,     /* @parm <t RFC_SAP_TXID>         */
 RFC_ERROR_INFO_EX * error_info    /* @parm <t RFC_ERROR_INFO_EX>
			            * describing the error
				    */
 );

/* @func
 *
 *
 * see <f RfcTxStart> for more details
 *
 *
 */
RFC_RC SAP_API RfcTxPrepareToCommit(
 RFC_HANDLE          handle,       /* @parm connection handle.       */
 void *              ext_txid,     /* @parm External Transaction ID. */
 int                 ext_txid_len, /* @parm Length of the External
		   	            * Transaction ID, Max. 32 B.
			   	    */
 RFC_SAP_TXID        sap_txid,     /* @parm <t RFC_SAP_TXID>         */
 RFC_ERROR_INFO_EX * error_info    /* @parm <t RFC_ERROR_INFO_EX>
			            * describing the error
				    */
 );

/* @func
 *
 *
 * see <f RfcTxStart> for more details
 *
 *
 */
RFC_RC SAP_API RfcTxCommit(
 RFC_HANDLE          handle,       /* @parm connection handle.       */
 void *              ext_txid,     /* @parm External Transaction ID. */
 int                 ext_txid_len, /* @parm Length of the External
		   	            * Transaction ID, Max. 32 B.
			   	    */
 RFC_SAP_TXID        sap_txid,     /* @parm <t RFC_SAP_TXID>         */
 RFC_ERROR_INFO_EX * error_info    /* @parm <t RFC_ERROR_INFO_EX>
			            * describing the error
				    */
 );

/* @func
 *
 *
 * see <f RfcTxStart> for more details
 *
 *
 */
RFC_RC SAP_API RfcTxAbort(
 RFC_HANDLE          handle,       /* @parm connection handle.       */
 void *              ext_txid,     /* @parm External Transaction ID. */
 int                 ext_txid_len, /* @parm Length of the External
		   	            * Transaction ID, Max. 32 B.
			   	    */
 RFC_SAP_TXID        sap_txid,     /* @parm <t RFC_SAP_TXID>         */
 RFC_ERROR_INFO_EX * error_info    /* @parm <t RFC_ERROR_INFO_EX>
			            * describing the error
				    */
 );

/* @func
 *
 * uninstalls an RFC function.
 *
 *
 * @xref <f RfcInstallFunction>
 *
 */
RFC_RC SAP_API RfcUninstallFunction(
 RFC_FUNCTIONNAME    f_name        /* @parm Name of an RFC function
                                    * which should be unstalled.
                                    */
 );


/* @type RFC_GUI_START_PARAM |
 *
 * zero terminated string giving parameters which SAPGUI should be
 * started with (returned by <f RfcGuiPrepare> and should be used
 * in <f RfcGuiStart>.
 *
 */
#define RFC_GUI_START_PARAM_LN   512
typedef rfc_char_t RFC_GUI_START_PARAM[RFC_GUI_START_PARAM_LN+1];


/* @type RFC_GUI_RETURN_PARAM |
 *
 * zero terminated string giving information by SAPGUI after
 * being started. This information must be sent to R/3 via the
 * RFC library using <f RfcGuiReturn>.
 *
 */
#define RFC_GUI_RETURN_PARAM_LN   128
typedef rfc_char_t RFC_GUI_RETURN_PARAM[RFC_GUI_RETURN_PARAM_LN+1];


/* @type RFC_GUI_ERROR |
 *
 * zero terminated string giving error message by unsuccessful start
 * of SAPGUI. This error message must be sent to R/3 via the
 * RFC library using <f RfcGuiReturn>.
 *
 * If no error occurs this error message  must be an empty string
 * (string length 0).
 *
 */
#define RFC_GUI_ERROR_LN   128
typedef rfc_char_t RFC_GUI_ERROR[RFC_GUI_ERROR_LN+1];


/* @func
 *
 * Additional to the automatic start of SAPGUI at the connection open time,
 * the RFC library supports now 4 more functions for working with SAPGUI.
 *
 * Using these new functions, SAPGUI will be started when needed
 * and not necessary at the connection open time.
 *
 * RfcGuiCheck     checks whether SAPGUI is already running.
 *
 * RfcGuiPrepare   informs the RFC library about working with SAPGUI.
 *                 As output, it will return some parameters which SAPGUI
 *                 should be started with.
 *
 * RfcGuiStart     starts SAPGUI via the RFC library.
 *
 * RfcGuiReturn    sends the information returned by SAPGUI after being
 *                 started to R/3.
 *
 *
 * Start SAPGUI on the same computer as RFC client (locale):
 *
 * If you only want to start SAPGUI on the same computer where your RFC
 * client is running but not at the OPEN-time, you only need to use the
 * function RfcGuiPrepare. In this case, the parameter remote_gui must
 * be set to 0.
 *
 *
 * Start SAPGUI not on the same computer as RFC client (remote):
 *
 * There are 2 different ways how to work with this.
 *
 *
 * 1. Start SAPGUI in your responsibility:
 *
 *    If your RFC client can send and receive with another program
 *    running on the remote computer, it must issue RfcGuiPrepare to get the
 *    parameters needed for starting SAPGUI. The program on the remote
 *    computer has then to start SAPGUI with these parameters using the
 *    function RfcGuiStart.
 *
 *    After SAPGUI is started, it will return some information to R/3, the
 *    program on the remote computer must send this information to your
 *    RFC client and it must then send this information to R/3 using
 *    RfcGuiReturn.
 *
 *    Using this functionality, you are responsible for sending
 *    START-parameters returned by RfcGuiPrepare to the remote computer
 *    for starting SAPGUI and sending information returned by SAPGUI after
 *    being started from the remote computer to R/3 via your RFC client and
 *    the RFC library.
 *
 *
 * 2. Using any RFC server in registered mode on the remote computer:
 *
 *    After starting an RFC server such as rfcexec or srfcserv delivered
 *    with the RFC SDK in registered mode on the remote computer, your RFC
 *    client only needs to issue RfcGuiPrepare with the information about the SAP
 *    gateway and the program ID which the RFC server is registered with.
 *    The parameter register_info contains this information must be in the
 *    form as defined in the following example:
 *
 *    GWHOST=hs0311 GWSERV=sapgw53 PROGID=hw1145.srfcserv
 *
 *    The RFC library on both computers will start SAPGUI on the remote
 *    computer as rewuired.
 *
 *    At the end of your work, your RFC client can use <f RfcCancelRegisterServer>
 *    to normally end the registered RFC server on the remote computer.
 *
 *
 * @xref <f RfcGuiCheck>, <f RfcGuiStart>, <f RfcGuiReturn>
 *
 */
RFC_RC SAP_API RfcGuiPrepare(
 RFC_HANDLE            handle,     /* @parm connection handle               */
 RFC_INT               remote_gui, /* @parm 0/1: locale/remote SAPGUI       */
 RFC_INT               debug,      /* @parm 0/1: without/with ABAP-debugger */
 RFC_INT               sapgui,     /* @parm 1/2: without/with invisible
                                    * SAPGUI after each RFC call
                                    */
 rfc_char_t                * register_info, /* @parm Info about the registered server
                                       * on the remote computer.
                                       * If using a registered server, the
                                       * parameter remote_gui must be set to 1.
                                       * If not, this parameter must be NULL or
                                       * an empty string (length 0)
                                       */
 RFC_GUI_START_PARAM  start_param, /* @parm <t RFC_GUI_START_PARAM>         */
 RFC_ERROR_INFO_EX  * error_info   /* @parm <t RFC_ERROR_INFO_EX>           */
 );


/* @func
 *
 *
 * Check whether SAPGUI is already running for this RFC connection.
 *
 *
 * @rdesc
 * return values
 * @flag 0      | SAPGUI is not running
 * @flag 1      | SAPGUI is already running
 *
 */
int SAP_API RfcGuiCheck(
 RFC_HANDLE            handle      /* @parm connection handle               */
 );



/* @func
 *
 *
 * see <f RfcGuiPrepare> for more details
 *
 *
 */
void SAP_API RfcGuiStart(
 RFC_GUI_START_PARAM     start_param,    /* @parm <t RFC_GUI_START_PARAM>  */
 RFC_GUI_ERROR           error_message,  /* @parm <t RFC_GUI_ERROR>        */
 RFC_GUI_RETURN_PARAM    return_param    /* @parm <t RFC_GUI_RETURN_PARAM> */
 );


/* @func
 *
 *
 * see <f RfcGuiPrepare> for more details
 *
 *
 */
RFC_RC SAP_API RfcGuiReturn(
 RFC_HANDLE             handle,         /* @parm connection handle         */
 RFC_GUI_ERROR          error_message,  /* @parm <t RFC_GUI_ERROR>         */
 RFC_GUI_RETURN_PARAM   return_param,   /* @parm <t RFC_GUI_RETURN_PARAM>  */
 RFC_ERROR_INFO_EX    * error_info      /* @parm <t RFC_ERROR_INFO_EX>     */
 );


/*------------------------------------------------------------------------------
 *
 * @func
 *
 * Check RFC_HANDLE for validity.
 *
 * RFC_ERROR_INFO may be NULL.
 *
 *
 * @rdesc
 * <e RFC_RC.RFC_OK>                Handle is valid
 *
 * <e RFC_RC.RFC_INVALID_HANDLE>    Handle is invalid
 *
 * <e RFC_RC.RFC_CLOSED>            Handle is valid but already disconected
 *                                  RfcClose must be called to release resources
 * <e RFC_RC.RFC_NOT_OWNER>         Handle used in Wrong Thread
 *                                  (when this check is enabled)
 *
 * @ex
 * This comes handy in
 *
 *  assert(RFC_OK == RfcIsValidHandle(hRfc, NULL));
 *
 *  or in
 *
 *  assert(  (RFC_OK     == RfcIsValidHandle(hRfc, NULL)
 *        || (RFC_CLOSED == RfcIsValidHandle(hRfc, NULL)
 *        );
 *
 *
 *
 *
 *-----------------------------------------------------------------------------*/
RFC_RC SAP_API RfcIsValidHandle(
 RFC_HANDLE             handle,         /* @parm connection handle         */
 RFC_ERROR_INFO_EX    * error_info      /* @parm <t RFC_ERROR_INFO_EX> A Description of
                                         * from the error
                                         */
 );

/* @func
 * Let the RFC lib check it's internal structures for integrity.
 *
 * <p error_info> may be NULL. This comes handy in
 *  assert(RFC_OK == RfcHealthCheck(NULL));
 *
 * @rvalue  RFC_OK       | No errors detected
 * @rvalue  RFC_FAILURE  | Some systems area are damaged
 */
RFC_RC SAP_API RfcHealthCheck(
 RFC_ERROR_INFO_EX    * error_info      /* @parm <t RFC_ERROR_INFO_EX> A Description of
                                         * from the error
                                         */
 );

/*---------------------------------------------------------------------
 * @enum RFC_CTL |
 *
 *   Possibly control request values for <f RfcIoCtl> function.
 *
 *---------------------------------------------------------------------*/
/* @emem any error has been occured */
typedef enum {
    RFC_CTL_NONE = 0,   /* @emem NOP */
    RFC_CTL_SET_MODNO,  /* @emem [in] int, internal use, sapgui only,
			 * pass emode to the rfc handler
			 */
    RFC_CTL_GET_MODNO,  /* @emem [out] int, internal use, sapgui only,
			 * retrieve emode modus
			 */
    RFC_CTL_GET_TRACE,  /* @emem [out] TraceOn, C(1), ' ' or 'X' */
    RFC_CTL_GET_CARRIER,/* @emem [out] RFC_HANDLE,
			 * in tRFC servers, retrieves handle from connection
			 * that transport the calls
			 */
    RFC_CTL_SET_DELTA_COMPACT, /* @emem [in] int, turns on the recycling
				* delta manager handles.
				* handy on very long running session
				*/
    RFC_CTL_SET_CRM_MODE, /* @emem [in] sets the OLE-DB Provider's converter in
			   * SAP-CRM mode
			   */
    RFC_CTL_GET_MEM_STATS,/* @emem [out] RFC_INT[4], retrieves memory usage statistics
			   * ((RFC_INT*)pOut)   = current number of allocated blocks
			   * ((RFC_INT*)pOut)+1 = current number of allocated bytes
			   * ((RFC_INT*)pOut)+2 = max number of allocated blocks so far
			   * ((RFC_INT*)pOut)+3 = max number of allocated bytes so far
			   */

    RFC_CTL_MAX /*@emem unsed end of enum marker */
} RFC_CTL;



/* @func
 * general purpose helper function to retrieve internal information
 *
 * <p handle> handle from which the information is requested.
 * For connection independant information, this may be RFC_HANDLE_NULL.
 * For connection dependant infromation, this must be a valid handle.
 *
 * <p ctlReq> the information to retrieve.
 *
 * use the following call:
 *
 *
 * { RFC_HANDLE hRfcCarrier;
 *   RFC_RC = RfcIoCtl( hRfctRfcCall, RFC_CTL_GET_CARRIER, 0, NULL, sizeof(hRfcCarrier), &hRfcCarrier);
 * }
 * { RFC_INT fOn = 1;
 *   RFC_RC = RfcIoCtl(  RFC_HANDLE_NULL, RFC_CTL_SET_DELTA_COMPACT, sizeof(fOn), &fOn, 0, NULL);
 * }
 * { RFC_INT rngStats[4];
 *   RFC_RC = RfcIoCtl(  RFC_HANDLE_NULL, RFC_CTL_GET_MEM_STATS, 0, NULL, sizeof(rngStats), rngStats);
 * }
 *
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> No errors detected
 * <e RFC_RC.Failure> Not implemented or not available
 */
RFC_RC SAP_API RfcIoCtl 
( 
 RFC_HANDLE     handle,     /* @parm handle of rfc connection */
 RFC_CTL        ctlReq,     /* @parm information to retrieve */
 int            sizeofpIn,  /* @parm size of area pointed by pIN */
 void    *      pIn,        /* @parm pointer to input area */
 int            sizeofpOut, /* @parm size of output area pointed by pOut */
 void    *      pOut        /* @parm pointer to output area */
);

/*
 * For internal use only
 *
 */

int SAP_API RfcNormalizeTabOrFldName(char * szOrgName,
										int uOrgSize,
                				        char * szNewName,
										int uNewSize);

RFC_RC SAP_API RfcCreateLocalTransID(RFC_TID tid);
RFC_RC SAP_API RfcGuidToTid(SAP_UUID *pGuid, RFC_TID Tid);
RFC_RC SAP_API RfcTidToGuid(RFC_TID Tid, SAP_UUID *pGuid);


#ifdef SAPGUI
/* ------------------------------------------------------------------
 *  functions must only be used by 'sapgui'
 *	and therefore not released to customers
 * ----------------------------------------------------------------*/
#include "saprfctr.h"
#endif /*SAPGUI*/


/*------------------------------------------------------
 *
 * @func
 *
 * Function to allocate memory for a <f RFC_STRING>.
 * The new allocated memory  will fullfilled with nulls. The space for terminating zero
 * will be allocated too.
 *
 * For better understanding you can consider RFC_STRING like  typedef unsigned char* RFC_STRING.
 * The standard C routines like strleng, strcpy etc. could be used to handle <t RFC_STRING>'s.
 *
 *
 * Warning:
 *
 *          1. Please use only this function for allocating
 *             of memory for RFC_STRING's.
 *             Do not use the functions malloc, realloc, free
 *             or C++ operators new/delete for string's memory management.
 *             You have to free strings created by this function only using <f RfcFreeString>.
 *
 *          2. Be carefully  strlen(RfcAllocString(x)) returns always 0.
 *
 * @rdesc
 *
 * a RFC_STRING if no error occurred
 *
 * NULL if call fails.
 *
 *
 * @xref <f RFC_STRING>, <f RfcFreeString>.
 *
 *
 * @ex  Using RfcAllocString |
 *
 *  RFC_STRING rfc_str = RfcAllocString (20);
 *
 *  strcpy (rfc_str, "Hello world");
 *  printf ("The length of rfc_str is: %d, the content of rfc_str is: %s", strlen(rfc_str), rfc_str);
 *
 *  RfcFreeString (rfc_str);
 *
 *------------------------------------------------------*/
RFC_STRING SAP_API RfcAllocString
(
  unsigned int  size    /* @parm  size of string. The allocated memory have size
                         * and is fullfilled with nulls.
						 * The call RfcAllocString(0) allocs 1 byte memory
						 * and fullfield it with null.
                         */

);




/*------------------------------------------------------
 * @func
 *
 * Function to free a <t RFC_STRING> allocated by <f RfcAllocString>.
 * Warning: Please use only this function for freeing
 *          of memory for RFC_STRING's allocated by <f RfcAllocString>.
 *          Do not use the functions malloc, realloc, free
 *          or C++ operators new/delete for string's memory managment.
 *          You have to use this function only for strings allocated by <f RfcAllocString>.
 *
 * @rdesc
 *
 * No returns
 *
 * @xref <f RFC_STRING>, <f RfcAllocString>.
 *
 *------------------------------------------------------*/
void SAP_API RfcFreeString
(
 RFC_STRING string /* @parm RFC_STRING allocated previosly by <f RfcAllocString>.
				    *
                    * If string is NULL, the function simply returns.
                    */
);






/*------------------------------------------------------
 * @func
 *
 * Function to convert 2 Byte ISO-Language in 1-byte SAP-Language.
 *
 *
 * @rdesc
 *
 * 0 OK
 * 1 ISO-language not found
 *
 * @xref <f RfcOpenEx>, <f RfcSapLangToIso>.
 *
 *------------------------------------------------------*/
int SAP_API RfcIsoLangToSap
(
 rfc_char_t	*isoLang,	/* @parm String ISO language */
 rfc_char_t	*sapLang	/* @parm String SAP language */
);




/*------------------------------------------------------
 * @func
 *
 * Function to convert 1-byte SAP-Language into 2 Byte ISO-Language.
 *
 *
 * @rdesc
 *
 * always 0 (OK).
 *
 * If saplang could not be found, isoLang = isoLang & blank.
 *
 * @xref <f RfcOpenEx>, <f RfcIsoLangToSap>.
 *
 *------------------------------------------------------*/
int SAP_API RfcSapLangToIso
(
 rfc_char_t		*sapLang,	/* @parm String ISO language */
 rfc_char_t		*isoLang	/* @parm String SAP language */
);




/*-------------------------------------------------------------------------
 *
 * @enum RFC_CANCEL_TYPE |
 * cancel mode to be set by <f RfcCancel> API.
 *
 * @xref <f RfcCancel>.
 *
 *--------------------------------------------------------------------------*/
typedef enum
{
 RFC_CANCEL_HARD = 0, /* @emem Not supported */

 RFC_CANCEL_SOFT = 1, /* @emem Not supported */

 RFC_CANCEL_CLOSE = 2 /* @emem Break the already running rfc function on ABAP-Side,
                       *        destroy user context on the ABAP-Side and
                       *         close the rfc connection.
                       */

} RFC_CANCEL_TYPE;

/*-----------------------------------------------------------------------
 * @func
 *
 * Function to cancel a RfcCall. This function can be used only by a rfc client running in
 * non blocking RfcCall mode (s. <f RfcListen>). The function itself is a blocking rfc call.
 *
 * If you try to cancel a rfc function running on any external server, the call fails
 * and server exits.
 *
 *
 * @rdesc
 *
 *    RFC_CANCELED if current running ABAP-Funktion could be canceled.
 *                            If parameter <e RFC_CANCEL_TYPE.RFC_CANCEL_SOFT>
 *                            was used, the user context in abap function was hold.
 *                            Rfc connection is not closed.
 *
 *    RFC_CLOSED   After canceling the current running ABAP-Funktion,
 *                            the rfc connection have been closed. This is the case if
 *                            the second input parameter is <e RFC_CANCEL_TYPE.RFC_CANCEL_CLOSE>
 *                            or partner closed rfc connection.
 *
 *    RFC_FAILURE if any error has occured.
 *
 *
 * @xref <f RFC_CANCEL_TYPE>, <f RfcCall>, <f RfcListen>, <f RfcReceive>.
 *
 * @ex using RfcCancel |
 *
 * RFC_RC rc;
 * RFC_HANDLE handle;
 *
 * ...
 *
 * rc = RfcCall( handle, .... );
 *
 * do
 * {
 *  rc = RfcListen( handle );
 *  if( rc == RFC_RETRY )
 *  {
 *   if (somethingHappened() == TRUE)
 *     rc = RfcCancel ( handle, RFC_CANCEL_SOFT);
 *
 *	  ....
 *
 *	doSomething();
 *  }
 * } while ( rc == RFC_RETRY );
 *
 * if( rc == RFC_OK )
 *  rc = RfcReceive( handle, .... );
 * ...
 *
 *----------------------------------------------------------------------*/
RFC_RC SAP_API RfcCancel
(
 RFC_HANDLE             rfc_handle,          /* @parm connection handle. */
 RFC_CANCEL_TYPE        cancel_type          /* @parm cancel type. */
);




/*-------------------------------------------------------------------------
 * @enum RFC_CONVERSION_FLAG |
 *
 * Flag to turn off/on the conversion for <f RFC_CHAR>'s between own and partners codepage.
 * Will be used by <f RfcCharConversion> .
 *
 * @xref <f RfcCharConversion>
 *
 *--------------------------------------------------------------------------*/
typedef enum
{
  On  = 0,	/* @emem Turn conversion for <f RFC_CHAR> between own and partners codepage on. */
  Off = 1	/* @emem Turn conversion for <f RFC_CHAR> between own and partners codepage off. */
} RFC_CONVERSION_FLAG;





/*------------------------------------------------------
 * @func
 * Function to turn On/Off the converton for <f RFC_CHAR> beetween own and partners codepage.
 *
 * If flag <f RFC_CONVERSION_FLAG> is set to Off, the RFC_CHAR data will be
 * handled like row data.
 * 	
 * Attention: This feature is to be used only in exception situations
 * if using of <f RFC_BYTE> is impossible.
 *
 * Important: this feature is only available from 4.6C release onwards.
 *
 * Warning: If both client and server call this API, the incoming data
 *          will be handled according the call made locally. This feature can be used,
 *          to create follow situation: Server does not convert incoming data and client
 *          does. To allow this the server should call RfcCharConversion (handle, Off) and the
 *          client should call RfcCharConversion (handle, On).
 *
 *
 * @rdesc
 *
 * <e RFC_RC.RFC_OK> if successfull.
 *
 * <e RFC_RC.RFC_FAILURE> if any error has occured.
 *
 *
 * @xref <f RFC_CHAR>, <f RfcSetCodePage>, <f RFC_CONVERSION_FLAG>.
 *
 *------------------------------------------------------*/
RFC_RC SAP_API RfcCharConversion
(
   RFC_HANDLE           rfc_handle, /* @parm RFC handle */
   RFC_CONVERSION_FLAG  flag        /* @parm Flag to turn the conversion of char's on or off.
			       	     *
				     *  If <e RFC_CONVERSION_FLAG.Off>, RFC_CHAR will be
                                     *  managed like binary data type.
				     *
				     *  If <e RFC_CONVERSION_FLAG.On>, RFC_CHAR will be
				     *  managed as usuall.
                                     */
);





/*-----------------------------------------------------------------------------------------
 * @func
 *
 * Copies the received cookie version 2 ticket into given user allocated memory area.
 * The request for the ticket generation has to be done by <f RfcOpenEx> function with flag GETSSO2
 * before calling this API.
 *
 * Features:
 *
 * This api can be called only once after <f RfcOpenEx>.
 * The second call of this api for the same handle fails.
 *
 * The call without request in <f RfcOpenEx> fails.
 *
 * Warning: The caller is responsible for memory allocation/deallocation for ticket.
 * The memory is to be allocated before api is called.
 *
 * @rdesc
 * <e RFC_RC.RFC_OK> if call succesfull
 *
 * <e RFC_RC.RFC_INVALID_HANDLE> if handle is invalid
 *
 * <e RFC_RC.RFC_FAILURE> if call fails.
 *
 *
 * @xref <f RfcOpenEx>.
 *-------------------------------------------------------------------------------------------*/
RFC_RC SAP_API RfcGetTicket
(
 RFC_HANDLE handle,	/*@parm rfc handle */
 RFC_CHAR*  ticket      /*@parm  Pointer on memory area allocated for ticket by caller. Must be long enough to receive the ticket*/
);


/*------------------------------------------------------
 *
 * @func
 *
 *  Function to allocate, reallocate or free a <t RFC_XSTRING>.
 *
 *  @rdesc
 *  <e RFC_RC.RFC_OK> if no error occurred and
 *  <e RFC_RC.RFC_FAILURE> elsewhere.
 *
 *  @xref <f RFC_BLOB>
 *
 *------------------------------------------------------*/
RFC_RC SAP_API RfcResizeXString
(
 RFC_XSTRING    *xstring,
 RFC_INT        length
);




#ifdef __cplusplus
}
#endif /*cplusplus*/



#endif /* SAPRFC_H */
#endif /* RFC_H */
